"""Compute job classes for interacting with a Media DCR.

These classes are meant to be used indirectly through the MediaDcr wrapper class.
"""

import io
import json
import time
from typing import Any, Dict, List, Optional, Tuple
import zipfile
from typing_extensions import Self
from abc import ABC, abstractmethod
from decentriq_dcr_compiler.schemas.data_room import DataRoomComputeAction
from .api import (
    ComputeInsightsResult,
    OverlapStatisticsResult,
    ValidationReport,
    DataAttributesResult,
    LookalikeAudienceStatisticsResult,
    GetSeedAudiencesResult,
    GetCustomAudiencesResult,
    SeedAudience,
    CustomAudience,
)
from ..client import Client
from ..storage import Key
from .audiences import Audiences, AudienceType
from .audience_statistics_result import LalAudienceStatistics
from .validation_reports import ValidationReports


class ComputeJob(ABC):
    """
    Abstract base class for compute jobs in a Media DCR.
    """

    def __init__(
        self,
        dcr_id: str,
        action: DataRoomComputeAction,
        result_zip_file_name: str,
        client: Client,
    ):
        """
        Initialize a compute job.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `action`: The compute action to perform
        - `result_zip_file_name`: The name of the file in the zip to get the result from
        - `client`: The client instance for API communication
        """
        self.dcr_id = dcr_id
        self.client = client
        self.client_v2 = (
            client.client_v2
        )  # TODO: Remove this once we have a single client
        self.session = client.create_session_v2()
        self.action = action
        self.result_zip_file_name = result_zip_file_name
        self.job_id = None

    def run(self) -> None:
        """
        Run the compute job.

        **Raises**:
        - `Exception`: If the computation has already been run
        """
        if self.job_id is not None:
            raise Exception("Computation has already been run")
        job = self.client_v2.create_job(
            self.dcr_id, self.action.model_dump(mode="json")
        )
        self.job_id = job["id"]

    def wait_for_completion(
        self, timeout: Optional[int] = None, sleep_interval: int = 1
    ) -> Self:
        """
        Wait for the compute job to complete.

        **Parameters**:
        - `timeout`: The maximum time to wait for the job to complete, in seconds
        - `sleep_interval`: The interval to wait between checks, in seconds

        **Returns**:
        - The compute job
        """
        start_time = time.time()
        while not self.is_complete():
            if timeout and time.time() - start_time > timeout:
                raise Exception("Computation did not complete within timeout")
            time.sleep(sleep_interval)
        return self

    def is_complete(self) -> bool:
        """
        Check if the compute job is complete.

        **Returns**:
        - True if the job is complete, False otherwise
        """
        if self.job_id is None:
            raise Exception("Computation has not been run")
        job_status = self.client_v2.get_job_status(self.job_id)
        return job_status["completed_at"] is not None

    @abstractmethod
    def result(self) -> Any:
        # Each job instance should implement this method and
        # return the appropriate type which represents the job result.
        pass

    def get_result_as_zipfile(self) -> zipfile.ZipFile:
        """Get the result of the compute job as a zipfile.ZipFile.

        **Returns**:
        - The content of the result file as a zipfile.ZipFile. If the job failed, an Exception will be raised.
        """
        if self.job_id is None:
            raise Exception(f"Job '{self.job_id}' has not been run")

        job_status = self.client_v2.get_job_status(self.job_id)
        if job_status["completed_at"] is None:
            raise Exception(f"Job '{self.job_id}' is not complete")

        job_tasks_view = self.client_v2.get_job_tasks(self.job_id)
        if job_status["status"] == "Failed":
            error_messages = []
            for task_view in job_tasks_view["tasks"].values():
                for selector, result_hash in task_view["results"].items():
                    if selector.startswith("error"):
                        if selector in task_view["completed"]:
                            result = self.download_result(result_hash)
                            result_bytes = result.read()
                            error_messages.append(
                                (task_view["name"], result_bytes.decode("utf-8"))
                            )
            formatted_error_messages = "\n".join(
                [
                    f"{task_name}: {error_message}"
                    for task_name, error_message in error_messages
                ]
            )
            raise Exception(
                f"Job '{self.job_id}' failed with errors: {formatted_error_messages}"
            )
        elif job_status["status"] == "Success":
            task_result_hash = job_tasks_view["tasks"][job_tasks_view["target_task"]][
                "results"
            ]["success"]
            result = self.download_result(task_result_hash)
            result_bytes = result.read()
            zip = zipfile.ZipFile(io.BytesIO(result_bytes), "r")
            return zip
        else:
            raise Exception(
                f"Job '{self.job_id}' has an unexpected status: {job_status['status']}"
            )

    def get_result(self) -> bytes:
        """
        Get the result of the compute job.

        **Returns**:
        - The content of the result file as bytes or raises an exception if the job failed
        """
        zip = self.get_result_as_zipfile()
        if self.result_zip_file_name in zip.namelist():
            return zip.read(self.result_zip_file_name)
        else:
            raise Exception(f"Did not find {self.result_zip_file_name} in zip")

    def download_result(self, task_result_hash: str) -> io.RawIOBase:
        """
        Download the result of the compute job.

        **Parameters**:
        - `task_result_hash`: The hash of the task result to download

        **Returns**:
        - The result of the compute job
        """
        manifest_hash, encryption_key = self._get_result_encryption_key(
            task_result_hash
        )
        key = Key(material=encryption_key)
        return self.client.download_job_result(
            self.job_id, manifest_hash, task_result_hash, key
        )

    def _get_result_encryption_key(self, task_result_hash: str) -> Tuple[str, bytes]:
        """
        Get the encryption key for the result of the compute job.

        **Parameters**:
        - `task_result_hash`: The hash of the task result to get the encryption key for

        **Returns**:
        - The encryption key for the result
        """
        manifest_hash, encryption_key = (
            self.session.send_retrieve_result_encryption_key_request(
                self.job_id, task_result_hash
            )
        )
        return manifest_hash, encryption_key

    def _get_task_result_hash(self, task_name: str) -> str:
        """
        Get the hash of the task result for a given task name.

        **Parameters**:
        - `task_name`: The name of the task to get the result hash for

        **Returns**:
        - The hash of the task result
        """
        job_tasks = self.client_v2.get_job_tasks(self.job_id)
        for _id, task_properties in job_tasks["tasks"].items():
            if task_properties["name"] == task_name:
                return task_properties["results"]["success"]
        raise Exception(f"Did not find task result hash for task {task_name}")


class MediaInsightsJob(ComputeJob):
    """
    A compute job for retrieving insights from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaInsightsJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "computeInsights": {}}
        )
        super().__init__(dcr_id, action, "segments.json", client)

    def result(self) -> ComputeInsightsResult:
        """
        Get the result of the MediaInsightsJob.

        **Returns**:
        - The result of the MediaInsightsJob
        """
        result = self.get_result()
        return ComputeInsightsResult.model_validate_json(result)


class MediaBaseAudienceValidationReportsJob:
    """
    A class to encapsulate the various base audience validation report jobs for a DCR.

    This class does not inherit from ComputeJob because it encapsulates multiple jobs rather than being a single job itself.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaBaseAudienceValidationReportsJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        self.dcr_id = dcr_id
        self.client = client
        self.validation_jobs = {
            "matching": MediaMatchingValidationReportJob(dcr_id, client),
            "segments": MediaSegmentsValidationReportJob(dcr_id, client),
            "demographics": MediaDemographicsValidationReportJob(dcr_id, client),
        }

    def run(self) -> None:
        """
        Run all validation jobs.
        """
        [job.run() for job in self.validation_jobs.values()]

    def wait_for_completion(
        self, timeout: Optional[int] = None, sleep_interval: int = 1
    ) -> Self:
        """
        Wait for the compute job to complete.

        **Parameters**:
        - `timeout`: The maximum time to wait for the job to complete, in seconds
        - `sleep_interval`: The interval to wait between checks, in seconds

        **Returns**:
        - The compute job
        """
        start_time = time.time()
        while not all([job.is_complete() for job in self.validation_jobs.values()]):
            if timeout and time.time() - start_time > timeout:
                raise Exception("Computation did not complete within timeout")
            time.sleep(sleep_interval)
        return self

    def result(self) -> ValidationReports:
        """
        Get the result of the MediaBaseAudienceValidationReportsJob.

        **Returns**:
        - The result of the MediaBaseAudienceValidationReportsJob
        """
        return ValidationReports(
            matching=self.validation_jobs["matching"].result(),
            segments=self.validation_jobs["segments"].result(),
            demographics=self.validation_jobs["demographics"].result(),
        )


class ValidationReportJob(ComputeJob):
    """
    A compute job for retrieving validation reports from a Media DCR.
    """

    def __init__(self, dcr_id: str, action: DataRoomComputeAction, client: Client):
        """
        Initialize a ValidationReportJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `action`: The compute action to perform
        - `client`: The client instance for API communication
        """
        super().__init__(dcr_id, action, "validation-report.json", client)

    def get_validation_report(self) -> ValidationReport:
        """
        Get the validation report for the given task name.

        **Parameters**:
        - `task_name`: The name of the task to get the validation report for

        **Returns**:
        - The validation report
        """
        result = self.get_result()
        return ValidationReport.model_validate_json(result)


class MediaMatchingValidationReportJob(ValidationReportJob):
    """
    A compute job for retrieving matching validation reports from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaMatchingValidationReportJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getMatchingValidationReport": {}}
        )
        super().__init__(dcr_id, action, client)

    def result(self) -> ValidationReport:
        """
        Get the result of the MediaMatchingValidationReportJob.

        **Returns**:
        - The result of the MediaMatchingValidationReportJob
        """
        return self.get_validation_report()


class MediaSegmentsValidationReportJob(ValidationReportJob):
    """
    A compute job for retrieving segments validation reports from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaSegmentsValidationReportJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getSegmentsValidationReport": {}}
        )
        super().__init__(dcr_id, action, client)

    def result(self) -> ValidationReport:
        """
        Get the result of the MediaSegmentsValidationReportJob.

        **Returns**:
        - The result of the MediaSegmentsValidationReportJob
        """
        return self.get_validation_report()


class MediaDemographicsValidationReportJob(ValidationReportJob):
    """
    A compute job for retrieving demographics validation reports from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaDemographicsValidationReportJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getDemographicsValidationReport": {}}
        )
        super().__init__(dcr_id, action, client)

    def result(self) -> ValidationReport:
        """
        Get the result of the MediaDemographicsValidationReportJob.

        **Returns**:
        - The result of the MediaDemographicsValidationReportJob
        """
        return self.get_validation_report()


class MediaAudiencesValidationReportJob(ValidationReportJob):
    """
    A compute job for retrieving audiences validation reports from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaAudiencesValidationReportJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getAudiencesValidationReport": {}}
        )
        super().__init__(dcr_id, action, client)

    def result(self) -> ValidationReport:
        """
        Get the result of the MediaAudiencesValidationReportJob.

        **Returns**:
        - The result of the MediaAudiencesValidationReportJob
        """
        return self.get_validation_report()


class MediaOverlapStatisticsJob(ComputeJob):
    """
    A compute job for retrieving overlap statistics from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaOverlapStatisticsJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "computeOverlapStatistics": {}}
        )
        super().__init__(dcr_id, action, "overlap.json", client)

    def result(self) -> OverlapStatisticsResult:
        """
        Get the result of the MediaOverlapStatisticsJob.

        **Returns**:
        - The result of the MediaOverlapStatisticsJob
        """
        result = self.get_result()
        return OverlapStatisticsResult.model_validate_json(result)


class MediaDataAttributesJob(ComputeJob):
    """
    A compute job for retrieving data attributes from a Media DCR.
    """

    def __init__(self, dcr_id: str, client: Client):
        """
        Initialize a MediaDataAttributesJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `client`: The client instance for API communication
        """
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getDataAttributes": {}}
        )
        super().__init__(dcr_id, action, "attributes.json", client)

    def result(self) -> DataAttributesResult:
        """
        Get the result of the MediaDataAttributesJob.

        **Returns**:
        - The result of the MediaDataAttributesJob
        """
        result = self.get_result()
        return DataAttributesResult.model_validate_json(result)


class MediaLookalikeAudienceStatisticsJob(ComputeJob):
    """
    A compute job for retrieving lookalike audience statistics from a Media DCR.
    """

    def __init__(self, dcr_id: str, audience: AudienceType, client: Client):
        """
        Initialize a MediaLookalikeAudienceStatisticsJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `audience`: The audience to get the statistics for
        - `client`: The client instance for API communication
        """
        if isinstance(audience, SeedAudience):
            audience_ref = {
                "kind": audience.kind.value,
                "audienceType": audience.audienceType,
            }
        elif isinstance(audience, CustomAudience):
            audience_ref = {
                "kind": "CUSTOM",
                "id": audience.id,
            }
        else:
            raise Exception(
                f"Cannot compute lookalike statistics for audience of unknown type: {type(audience)}"
            )
        action = DataRoomComputeAction.model_validate(
            {
                "kind": "media",
                "getLookalikeAudienceStatistics": {"audienceRef": audience_ref},
            }
        )
        super().__init__(dcr_id, action, "lookalike_audience.json", client)

    def result(self) -> LalAudienceStatistics:
        """
        Get the result of the MediaLookalikeAudienceStatisticsJob.

        **Returns**:
        - Lookalike statistics as a LalAudienceStatistics object
        """
        result = self.get_result()
        lal_statistics_result = LookalikeAudienceStatisticsResult.model_validate_json(
            result
        )
        formatted_result = LalAudienceStatistics.from_dict(
            lal_statistics_result.model_dump()
        )
        return formatted_result


class MediaGetCustomAudiencesJob(ComputeJob):
    """
    A compute job for retrieving custom audiences from a Media DCR.
    """

    def __init__(self, dcr_id: str, user: str, client: Client):
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getCustomAudiences": {"user": user}}
        )
        super().__init__(dcr_id, action, "audiences.json", client)

    def result(self) -> GetCustomAudiencesResult:
        result = self.get_result()
        return GetCustomAudiencesResult.model_validate_json(result.decode("utf-8"))


class MediaGetSeedAudiencesJob(ComputeJob):
    """
    A compute job for retrieving seed audiences from a Media DCR.
    """

    def __init__(self, dcr_id: str, user: str, client: Client):
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "getSeedAudiences": {"user": user}}
        )
        super().__init__(dcr_id, action, "audiences.json", client)

    def result(self) -> GetSeedAudiencesResult:
        result = self.get_result()
        return GetSeedAudiencesResult.model_validate_json(result.decode("utf-8"))


class MediaGetAudiencesJob(ComputeJob):
    """
    A compute job for retrieving all audiences from a Media DCR.
    """

    def __init__(self, dcr_id: str, user: str, client: Client, participant_id: str):
        """
        Initialize a MediaGetAudiencesJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `user`: The user to get the audiences for
        - `client`: The client instance for API communication
        """
        self._custom = MediaGetCustomAudiencesJob(dcr_id, user, client)
        self._seed = MediaGetSeedAudiencesJob(dcr_id, user, client)
        self._participant_id = participant_id
        self._user_email = user

    def run(self):
        """
        Run the compute job.

        **Raises**:
        - `Exception`: If the computation has already been run
        """
        self._custom.run()
        self._seed.run()

    def wait_for_completion(
        self, timeout: Optional[int] = None, sleep_interval: int = 1
    ) -> Self:
        """
        Wait for the compute job to complete.

        **Parameters**:
        - `timeout`: The maximum time to wait for the job to complete, in seconds
        - `sleep_interval`: The interval to wait between checks, in seconds

        **Returns**:
        - The compute job
        """
        self._custom.wait_for_completion(timeout=timeout, sleep_interval=sleep_interval)
        self._seed.wait_for_completion(timeout=timeout, sleep_interval=sleep_interval)
        return self

    def is_complete(self) -> bool:
        """
        Check if the compute job is complete.

        **Returns**:
        - True if the job is complete, False otherwise
        """
        return self._custom.is_complete() and self._seed.is_complete()

    def result(self) -> Audiences:
        """
        Get the result of the MediaGetAudiencesJob.

        **Returns**:
        - The result of the MediaGetAudiencesJob
        """
        seed_result = self._seed.result()
        custom_result = self._custom.result()
        all_custom_audiences = custom_result.audiences
        # The backend will return all custom audiences that are either shared
        # with the user or that are dependencies of such audiences.
        # This is required because in the UI we also want to display information
        # about how some shared audience is constructed (e.g. the names of which
        # audiences are combined in a rule-based audience).
        # In the SDK this information is not required and for that reason these
        # audiences are removed here since they could not be exported by the
        # current user.
        accessible_custom_audiences = []
        for audience in all_custom_audiences:
            shared_with_participants = audience.sharedWith or []
            is_owner = audience.createdBy == self._user_email
            is_shared = self._participant_id in shared_with_participants
            if is_owner or is_shared:
                accessible_custom_audiences.append(audience)
        audiences: list[AudienceType] = (
            accessible_custom_audiences + seed_result.audiences
        )
        return Audiences(audiences)


class MediaEstimateAudienceSizeJob(ComputeJob):
    """
    A compute job for estimating audience size in a Media DCR.
    """

    def __init__(self, dcr_id: str, audience: AudienceType, client: Client):
        """
        Initialize a MediaEstimateAudienceSizeJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `audience`: The audience to get the size for
        - `client`: The client instance for API communication
        """
        self.is_lookalike = Audiences.is_lookalike_audience(audience)
        if isinstance(audience, SeedAudience):
            audience_ref = {
                "kind": audience.kind.value,
                "audienceType": audience.audienceType,
            }
        elif isinstance(audience, CustomAudience):
            audience_ref = {
                "kind": "CUSTOM",
                "id": audience.id,
            }

        action = DataRoomComputeAction.model_validate(
            {
                "kind": "media",
                "estimateAudienceSize": {
                    "audience": {"kind": "REF", "value": audience_ref}
                },
            }
        )
        super().__init__(dcr_id, action, "audience_size.json", client)

    def result(self) -> int:
        """
        Get the result of the MediaEstimateAudienceSizeJob.

        **Returns**:
        - The result of the MediaEstimateAudienceSizeJob
        """
        result = self.get_result()
        audience_size_result = json.loads(result)
        return int(audience_size_result["audience_size"])


class MediaAudienceUserListJob(ComputeJob):
    """
    A compute job for retrieving audience user lists from a Media DCR.
    """

    def __init__(self, dcr_id: str, audience: AudienceType, client: Client):
        """
        Initialize a MediaAudienceUserListJob instance.

        **Parameters**:
        - `dcr_id`: The identifier for the DCR
        - `audience`: The audience to get the user list for
        - `client`: The client instance for API communication
        """
        self.is_lookalike = Audiences.is_lookalike_audience(audience)
        if isinstance(audience, SeedAudience):
            audience_ref = {
                "kind": audience.kind.value,
                "audienceType": audience.audienceType,
            }
        elif isinstance(audience, CustomAudience):
            audience_ref = {
                "kind": "CUSTOM",
                "id": audience.id,
            }
        action = DataRoomComputeAction.model_validate(
            {
                "kind": "media",
                "getAudienceUserList": {"audienceRef": audience_ref},
            }
        )
        super().__init__(dcr_id, action, "audience_users.csv", client)

    def result(self) -> List[str]:
        """
        Get the result of the MediaAudienceUserListJob.

        **Returns**:
        - The result of the MediaAudienceUserListJob
        """
        result = self.get_result()
        result = result.decode("utf-8").splitlines()
        return [line for line in result if line]


class MediaModelQualityReportJob(ComputeJob):
    def __init__(self, dcr_id: str, client: Client):
        action = DataRoomComputeAction.model_validate(
            {"kind": "media", "retrieveModelQualityReport": {}}
        )
        super().__init__(dcr_id, action, "model_quality.json", client)

    def result(self) -> Dict[str, Any]:
        result = self.get_result()
        return json.loads(result.decode("utf-8"))



import os


def _use_tls():
    return False if os.getenv("DECENTRIQ_USE_TLS", "true") == "false" else True


def _disable_known_root_ca_check():
    return (
        True
        if os.getenv("_DECENTRIQ_UNSAFE_DISABLE_KNOWN_ROOT_CA_CHECK", "false") == "true"
        else False
    )


def _disable_dataset_acl_organization_check():
    return (
        True
        if os.getenv("DECENTRIQ_DISABLE_DATASET_ACL_ORGANIZATION_CHECK", "false") == "true"
        else False
    )


DECENTRIQ_CLIENT_ID = os.getenv(
    "DECENTRIQ_CLIENT_ID", "MHyVW112w7Ql95G96fn9rnLWkYuOLmdk"
)
DECENTRIQ_HOST = os.getenv("DECENTRIQ_HOST", "api.decentriq.com")
DECENTRIQ_PORT = int(os.getenv("DECENTRIQ_PORT", "443"))
DECENTRIQ_USE_TLS = _use_tls()
DECENTRIQ_REQUEST_RETRY_TOTAL = int(os.getenv("DECENTRIQ_REQUEST_RETRY_TOTAL", "3"))
DECENTRIQ_REQUEST_RETRY_BACKOFF_FACTOR = int(
    os.getenv("DECENTRIQ_REQUEST_RETRY_BACKOFF_FACTOR", "0")
)
DECENTRIQ_MRSIGNER_DRIVER_ATTESTATION_SPECIFICATION = os.getenv("DECENTRIQ_MRSIGNER_DRIVER_ATTESTATION_SPECIFICATION")
_DECENTRIQ_UNSAFE_DISABLE_KNOWN_ROOT_CA_CHECK = _disable_known_root_ca_check()
DECENTRIQ_DISABLE_DATASET_ACL_ORGANIZATION_CHECK = _disable_dataset_acl_organization_check()



from typing_extensions import Self
from .audience_definitions import (
    LookalikeAudienceDefinition,
    AudienceStatus,
    AudienceDefinitions,
    audience_id_from_name,
)


class LookalikeAudienceBuilder:
    """
    Builder for constructing lookalike audience definitions.
    """

    def __init__(
        self,
        *,
        name: str,
        reach: int,
        source_audience_name: str,
        audiences: AudienceDefinitions,
    ) -> None:
        """
        Initialise the lookalike audience builder.

        **Parameters**:
        - `name`: Name of the lookalike audience.
        - `reach`: The desired reach of the lookalike audience expressed as a percentage between 1-30.
        - `source_audience_name`: Name of the source audience that the lookalike audience will be built from.
        - `audiences`: Definitions of all existing audiences.
        """
        if not (0 <= reach <= 30):
            raise Exception(f"Reach value {reach} is not in range 0 to 30")

        self.name = name
        self.reach = reach
        self.source_audience_name = source_audience_name
        self.audiences = audiences
        self.exclude_seed_audience = False
        self.make_available_to_publisher = False

    def with_exclude_seed_audience(self) -> Self:
        """
        Exclude the seed audience from the lookalike audience.
        """
        self.exclude_seed_audience = True
        return self

    def with_make_available_to_publisher(self) -> Self:
        """
        Make the lookalike audience available to the publisher.
        """
        self.make_available_to_publisher = True
        return self

    def build(self) -> LookalikeAudienceDefinition:
        """
        Build the lookalike audience definition.
        """
        source_ref_id = audience_id_from_name(self.source_audience_name, self.audiences)
        return LookalikeAudienceDefinition(
            name=self.name,
            reach=self.reach,
            seed_audience_name=self.source_audience_name,
            source_ref_id=source_ref_id,
            exclude_seed_audience=self.exclude_seed_audience,
            status=(
                AudienceStatus.PUBLISHED
                if self.make_available_to_publisher
                else AudienceStatus.READY
            ),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from typing_extensions import Self

from ..attestation import enclave_specifications
from ..session import Session
from ..types import EnclaveSpecification
from .existing_builder import ExistingAnalyticsDcrBuilder
from .high_level_node import ComputationNode, DataNode
from .node_definitions import NodeDefinition
from .version import DATA_SCIENCE_DCR_SUPPORTED_VERSION

if TYPE_CHECKING:
    from ..client import Client


class AnalyticsDcrDefinition:
    """
    A class representing an Analytics DCR Definition.
    """

    def __init__(
        self,
        name: str,
        high_level: Dict[str, Any],
        enclave_specs: Optional[Dict[str, EnclaveSpecification]] = None,
    ) -> None:
        self.name = name
        self.high_level = high_level
        self.enclave_specs = enclave_specs

    def _get_high_level_representation(self) -> Dict[str, Any]:
        return self.high_level

    def _get_description(self) -> str:
        return self.high_level[DATA_SCIENCE_DCR_SUPPORTED_VERSION]["interactive"][
            "initialConfiguration"
        ]["description"]


class AnalyticsDcr:
    """
    A class representing an Analytics DCR.
    """

    def __init__(
        self,
        session: Session,
        dcr_id: str,
        high_level: Dict[str, str],
        nodes: List[NodeDefinition],
        *,
        client: Client,
    ):
        """
        Initialise an Analytics DCR.

        **Parameters**:
        - `session`: A `Session` object which can be used for communication with the enclave.
        - `dcr_id`: ID of the Analytics DCR.
        - `high_level`: High level representation of the Analytics DCR.
        - `nodes`: List of Data Node Definitions in the Analytics DCR.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        """
        self.client = client
        self.session = session
        self.high_level = high_level
        self.node_definitions = nodes
        self.id = dcr_id

    def get_node(self, name: str) -> Optional[Union[ComputationNode, DataNode]]:
        """
        Retrieve a node with the given name.

        **Parameters**:
        - `name`: Node name.
        """
        for node_def in self.node_definitions:
            if node_def.name == name:
                node = node_def.build(
                    dcr_id=self.id,
                    node_definition=node_def,
                    client=self.client,
                    session=self.session,
                )
                return node
        return None

    def retrieve_audit_log(self) -> str:
        """
        Retrieve the Analytics DCR audit log.
        """
        return self.session.retrieve_audit_log(self.id).log.decode("utf-8")

    def stop(self):
        """
        Stop the Analytics DCR.
        """
        self.session.stop_data_room(self.id)

    @classmethod
    def _from_existing(
        cls,
        dcr_id: str,
        *,
        client: Client,
        enclave_specs: Optional[List[EnclaveSpecification]] = None,
    ) -> Self:
        """
        Construct a Analytics DCR from an existing DCR with the given ID.

        **Parameters**:
        - `dcr_id`: Data Clean Room ID.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `enclave_specs`: Determines the types of enclaves that are supported by this Data Clean Room.
            If not specified, the latest enclave specifications will be used.
        """
        data_room_descriptions = client.get_data_room_descriptions()
        existing_data_room_description = [
            description
            for description in data_room_descriptions
            if description["id"] == dcr_id
        ]
        if len(existing_data_room_description) != 1:
            raise Exception(
                f"Unable to retrieve data room description for data room with ID {dcr_id}"
            )
        specs = enclave_specs if enclave_specs else enclave_specifications.all()
        session = client.create_session_from_data_room_description(
            existing_data_room_description[0], specs
        )
        existing_dcr_builder = ExistingAnalyticsDcrBuilder(dcr_id, client, session)
        cfg = existing_dcr_builder.get_configuration()
        dcr = cls(
            client=client,
            session=session,
            dcr_id=dcr_id,
            high_level=existing_dcr_builder.high_level,
            nodes=cfg.node_definitions,
        )
        return dcr

    def participants(self) -> List[str]:
        """
        Retrieve the participants of the Analytics DCR as a list.
        """
        keys = list(self.high_level.keys())
        if len(keys) != 1:
            raise Exception(
                f"Unable to extract Analytics DCR version. Expected a single top-level property indicating the DCR version."
            )
        dcr_version = keys[0]

        dcr = self.high_level[dcr_version]
        keys = list(dcr.keys())
        if len(keys) != 1:
            raise Exception(
                f"Unable to extract the interactivity type for the Analytics DCR. Expected a single top-level property indicating the interactivity type."
            )

        dcr_type = keys[0]
        if dcr_type == "interactive":
            participants = dcr[dcr_type]["initialConfiguration"]["participants"]
        elif dcr_type == "static":
            participants = dcr[dcr_type]["participants"]
        else:
            raise Exception(
                f'Unknown DCR type {dcr_type}. Expected "interactive" or "static" type.'
            )
        return [participant["user"] for participant in participants]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _IdFormat:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IdFormatEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IdFormat.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNHASHED_EMAIL: _IdFormat.ValueType  # 0
    HASHED_EMAIL_SHA256: _IdFormat.ValueType  # 1
    PHONE_NR: _IdFormat.ValueType  # 2
    HASHED_PHONE_NR_SHA256: _IdFormat.ValueType  # 3

class IdFormat(_IdFormat, metaclass=_IdFormatEnumTypeWrapper): ...

UNHASHED_EMAIL: IdFormat.ValueType  # 0
HASHED_EMAIL_SHA256: IdFormat.ValueType  # 1
PHONE_NR: IdFormat.ValueType  # 2
HASHED_PHONE_NR_SHA256: IdFormat.ValueType  # 3
global___IdFormat = IdFormat

class _Country:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _CountryEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Country.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN: _Country.ValueType  # 0
    """default value"""
    UNITED_ARAB_EMIRATES: _Country.ValueType  # 1
    AUSTRALIA: _Country.ValueType  # 2
    BELGIUM: _Country.ValueType  # 3
    BRAZIL: _Country.ValueType  # 4
    CANADA: _Country.ValueType  # 5
    GERMANY: _Country.ValueType  # 6
    EGYPT: _Country.ValueType  # 7
    SPAIN: _Country.ValueType  # 8
    FRANCE: _Country.ValueType  # 9
    UNITED_KINGDOM: _Country.ValueType  # 10
    INDIA: _Country.ValueType  # 11
    ITALY: _Country.ValueType  # 12
    JAPAN: _Country.ValueType  # 13
    MEXICO: _Country.ValueType  # 14
    NETHERLANDS: _Country.ValueType  # 15
    POLAND: _Country.ValueType  # 16
    SAUDI_ARABIA: _Country.ValueType  # 17
    SWEDEN: _Country.ValueType  # 18
    SINGAPORE: _Country.ValueType  # 19
    TURKEY: _Country.ValueType  # 20
    UNITED_STATES_OF_AMERICA: _Country.ValueType  # 21

class Country(_Country, metaclass=_CountryEnumTypeWrapper): ...

UNKNOWN: Country.ValueType  # 0
"""default value"""
UNITED_ARAB_EMIRATES: Country.ValueType  # 1
AUSTRALIA: Country.ValueType  # 2
BELGIUM: Country.ValueType  # 3
BRAZIL: Country.ValueType  # 4
CANADA: Country.ValueType  # 5
GERMANY: Country.ValueType  # 6
EGYPT: Country.ValueType  # 7
SPAIN: Country.ValueType  # 8
FRANCE: Country.ValueType  # 9
UNITED_KINGDOM: Country.ValueType  # 10
INDIA: Country.ValueType  # 11
ITALY: Country.ValueType  # 12
JAPAN: Country.ValueType  # 13
MEXICO: Country.ValueType  # 14
NETHERLANDS: Country.ValueType  # 15
POLAND: Country.ValueType  # 16
SAUDI_ARABIA: Country.ValueType  # 17
SWEDEN: Country.ValueType  # 18
SINGAPORE: Country.ValueType  # 19
TURKEY: Country.ValueType  # 20
UNITED_STATES_OF_AMERICA: Country.ValueType  # 21
global___Country = Country

@typing.final
class SingleFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name"]) -> None: ...

global___SingleFile = SingleFile

@typing.final
class RawFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___RawFile = RawFile

@typing.final
class ZipFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SINGLEFILE_FIELD_NUMBER: builtins.int
    @property
    def singleFile(self) -> global___SingleFile: ...
    def __init__(
        self,
        *,
        singleFile: global___SingleFile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["selection", b"selection", "singleFile", b"singleFile"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["selection", b"selection", "singleFile", b"singleFile"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["selection", b"selection"]) -> typing.Literal["singleFile"] | None: ...

global___ZipFile = ZipFile

@typing.final
class DataInput(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEPENDENCY_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    RAW_FIELD_NUMBER: builtins.int
    ZIP_FIELD_NUMBER: builtins.int
    dependency: builtins.str
    name: builtins.str
    @property
    def raw(self) -> global___RawFile: ...
    @property
    def zip(self) -> global___ZipFile: ...
    def __init__(
        self,
        *,
        dependency: builtins.str = ...,
        name: builtins.str = ...,
        raw: global___RawFile | None = ...,
        zip: global___ZipFile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["file", b"file", "raw", b"raw", "zip", b"zip"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dependency", b"dependency", "file", b"file", "name", b"name", "raw", b"raw", "zip", b"zip"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["file", b"file"]) -> typing.Literal["raw", "zip"] | None: ...

global___DataInput = DataInput

@typing.final
class UnHashedEmail(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___UnHashedEmail = UnHashedEmail

@typing.final
class HashedEmailSha256(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___HashedEmailSha256 = HashedEmailSha256

@typing.final
class TargetResource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONNECTIONID_FIELD_NUMBER: builtins.int
    ADVERTISERID_FIELD_NUMBER: builtins.int
    connectionId: builtins.str
    advertiserId: builtins.str
    def __init__(
        self,
        *,
        connectionId: builtins.str = ...,
        advertiserId: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["advertiserId", b"advertiserId", "connectionId", b"connectionId", "target", b"target"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["advertiserId", b"advertiserId", "connectionId", b"connectionId", "target", b"target"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["target", b"target"]) -> typing.Literal["connectionId", "advertiserId"] | None: ...

global___TargetResource = TargetResource

@typing.final
class AmazonAmcWorkerConfiguration(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INPUT_FIELD_NUMBER: builtins.int
    CREDENTIALSDEPENDENCY_FIELD_NUMBER: builtins.int
    TARGETRESOURCE_FIELD_NUMBER: builtins.int
    AUDIENCEID_FIELD_NUMBER: builtins.int
    AUDIENCENAME_FIELD_NUMBER: builtins.int
    AUDIENCEDESCRIPTION_FIELD_NUMBER: builtins.int
    COUNTRY_FIELD_NUMBER: builtins.int
    IDFORMAT_FIELD_NUMBER: builtins.int
    credentialsDependency: builtins.str
    audienceId: builtins.str
    audienceName: builtins.str
    audienceDescription: builtins.str
    country: global___Country.ValueType
    idFormat: global___IdFormat.ValueType
    @property
    def input(self) -> global___DataInput: ...
    @property
    def targetResource(self) -> global___TargetResource: ...
    def __init__(
        self,
        *,
        input: global___DataInput | None = ...,
        credentialsDependency: builtins.str = ...,
        targetResource: global___TargetResource | None = ...,
        audienceId: builtins.str = ...,
        audienceName: builtins.str = ...,
        audienceDescription: builtins.str | None = ...,
        country: global___Country.ValueType = ...,
        idFormat: global___IdFormat.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_audienceDescription", b"_audienceDescription", "audienceDescription", b"audienceDescription", "input", b"input", "targetResource", b"targetResource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_audienceDescription", b"_audienceDescription", "audienceDescription", b"audienceDescription", "audienceId", b"audienceId", "audienceName", b"audienceName", "country", b"country", "credentialsDependency", b"credentialsDependency", "idFormat", b"idFormat", "input", b"input", "targetResource", b"targetResource"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["_audienceDescription", b"_audienceDescription"]) -> typing.Literal["audienceDescription"] | None: ...

global___AmazonAmcWorkerConfiguration = AmazonAmcWorkerConfiguration



"""
.. include:: ../../decentriq_platform_docs/gcg_getting_started.md
___
"""

from . import (
    analytics,
    data_lab,
    legacy,
    session,
    types,
    authentication,
    data_connectors,
    ab_media,
    media,
)
from .attestation import EnclaveSpecifications, enclave_specifications
from .client import Client, Session, create_client, SecretStoreOptions
from .endorsement import Endorser
from .storage import Key
from .archv2 import SessionV2, Secret
from decentriq_dcr_compiler.schemas import Permission

__docformat__ = "restructuredtext"

__pdoc__ = {
    "api": False,
    "attestation": True,
    "authentication": True,
    "builders": False,
    "certs": False,
    "client": False,
    "compute": False,
    "config": False,
    "graphql": False,
    "helpers": False,
    "proto": False,
    "session": False,
    "storage": False,
    "types": False,
    "verification": False,
    "data_lab": True,
    "legacy": False,
    "decoder": False,
    "data_connectors": True,
    "ab_media": True,
    "Permission": False,
    "media": True,
}

__all__ = [
    "create_client",
    "Client",
    "Session",
    "enclave_specifications",
    "EnclaveSpecifications",
    "Key",
    "storage",
    "attestation",
    "types",
    "authentication",
    "session",
    "Endorser",
    "data_lab",
    "analytics",
    "data_connectors",
    "ab_media",
    "SessionV2",
    "Secret",
    "SecretStoreOptions",
    "Permission",
    "media",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: azure_blob_storage.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'azure_blob_storage.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x61zure_blob_storage.proto\x12\x12\x61zure_blob_storage\"\x0c\n\nImportRole\":\n\nExportRole\x12,\n\x05input\x18\x01 \x01(\x0b\x32\x1d.azure_blob_storage.SinkInput\"\x1a\n\nSingleFile\x12\x0c\n\x04name\x18\x01 \x01(\t\"\t\n\x07RawFile\"L\n\x07ZipFile\x12\x34\n\nsingleFile\x18\x01 \x01(\x0b\x32\x1e.azure_blob_storage.SingleFileH\x00\x42\x0b\n\tselection\"\x8d\x01\n\tSinkInput\x12\x12\n\ndependency\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12*\n\x03raw\x18\x03 \x01(\x0b\x32\x1b.azure_blob_storage.RawFileH\x00\x12*\n\x03zip\x18\x04 \x01(\x0b\x32\x1b.azure_blob_storage.ZipFileH\x00\x42\x06\n\x04\x66ile\"\xb8\x01\n#AzureBlobStorageWorkerConfiguration\x12\x1d\n\x15\x63redentialsDependency\x18\x01 \x01(\t\x12\x34\n\nimportRole\x18\x02 \x01(\x0b\x32\x1e.azure_blob_storage.ImportRoleH\x00\x12\x34\n\nexportRole\x18\x03 \x01(\x0b\x32\x1e.azure_blob_storage.ExportRoleH\x00\x42\x06\n\x04roleb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'azure_blob_storage_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_IMPORTROLE']._serialized_start=48
  _globals['_IMPORTROLE']._serialized_end=60
  _globals['_EXPORTROLE']._serialized_start=62
  _globals['_EXPORTROLE']._serialized_end=120
  _globals['_SINGLEFILE']._serialized_start=122
  _globals['_SINGLEFILE']._serialized_end=148
  _globals['_RAWFILE']._serialized_start=150
  _globals['_RAWFILE']._serialized_end=159
  _globals['_ZIPFILE']._serialized_start=161
  _globals['_ZIPFILE']._serialized_end=237
  _globals['_SINKINPUT']._serialized_start=240
  _globals['_SINKINPUT']._serialized_end=381
  _globals['_AZUREBLOBSTORAGEWORKERCONFIGURATION']._serialized_start=384
  _globals['_AZUREBLOBSTORAGEWORKERCONFIGURATION']._serialized_end=568
# @@protoc_insertion_point(module_scope)



"""Classes for managing different types of audiences."""

from typing import List, Union
from .api import SeedAudience, CustomAudience

AudienceType = Union[SeedAudience, CustomAudience]


class Audiences:
    """
    Class for managing and accessing different types of audiences.
    """

    def __init__(self, audiences: List[AudienceType]):
        """
        Initialize an Audiences instance.

        **Parameters**:
        - `audiences`: List of audience objects to manage
        """
        self.audiences = audiences

    def list(self) -> List[AudienceType]:
        """
        List all audiences.

        **Returns**:
        - List of all audiences
        """
        return self.audiences

    def get_seed_audience(self, audience_type: str) -> SeedAudience:
        """
        Get a seed audience by type.

        **Parameters**:
        - `audience_type`: Type of the seed audience to retrieve

        **Returns**:
        - Seed audience with the given type

        **Raises**:
        - `Exception`: If no seed audience is found with the given type or multiple seed audiences found with the given type
        """
        seed_audiences = [
            a
            for a in self.audiences
            if isinstance(a, SeedAudience) and a.audienceType == audience_type
        ]
        if len(seed_audiences) == 0:
            raise Exception(f"No seed audience found with type {audience_type}")
        if len(seed_audiences) > 1:
            raise Exception(f"Multiple seed audiences found with type {audience_type}")
        return seed_audiences[0]

    def get_custom_audience(self, name: str) -> CustomAudience:
        """
        Get a custom audience by name.

        **Parameters**:
        - `name`: Name of the custom audience to retrieve

        **Returns**:
        - Custom audience with the given name

        **Raises**:
        - `Exception`: If no custom audience is found or multiple custom audiences found with the given name
        """
        custom_audiences = [
            a
            for a in self.audiences
            if isinstance(a, CustomAudience) and a.name == name
        ]
        if len(custom_audiences) == 0:
            raise Exception(f"No custom audience found with name {name}")
        if len(custom_audiences) > 1:
            raise Exception(f"Multiple custom audiences found with name {name}")
        return custom_audiences[0]

    def is_lookalike_audience(audience: AudienceType) -> bool:
        """
        Check if an audience is a lookalike audience.

        **Parameters**:
        - `audience`: The audience to check

        **Returns**:
        - True if the audience is a lookalike audience, False otherwise

        **Raises**:
        - `Exception`: If the audience type is not supported
        """
        if isinstance(audience, SeedAudience):
            return False
        elif isinstance(audience, CustomAudience):
            return audience.definition.kind == "LOOKALIKE"
        else:
            raise Exception(f"Unsupported audience type: {type(audience)}")



from __future__ import annotations

import json

from typing import Any, Dict, TYPE_CHECKING, List, Union
from decentriq_dcr_compiler import compiler, ab_media as ab_media_compiler
from decentriq_dcr_compiler.schemas import AbMediaRequest
from .publisher_computations import (
    GetAudiencesForPublisherComputation,
    EstimateAudienceSizeForPublisherComputation,
    GetAudienceUserListForPublisherComputation,
    EstimateAudienceSizeForPublisherLalComputation,
    GetAudienceUserListForPublisherLalComputation,
    ComputeInsightsComputation,
    GetSegmentsValidationReport,
    GetMatchingValidationReport,
    GetDemographicsValidationReport,
)
from .advertiser_computations import ComputeOverlapStatistics
from .helper import (
    get_parameter_payloads,
    audience_depends_on_lookalike,
)
from .audience_definitions import (
    RuleBasedAudienceDefinition,
    LookalikeAudienceDefinition,
    AdvertiserAudienceDefinition,
)
from .request import Request
from ..storage import Key

if TYPE_CHECKING:
    from ..client import Client


class PublisherApi:
    """
    Provides the API for interacting with an Audience Builder DCR as a publisher.
    """

    def __init__(self, dcr: AbMediaDcr) -> None:
        self.dcr_id = dcr.id
        self.session = dcr.session
        self.client = dcr.client
        self.high_level_dcr = dcr.high_level
        self.features = dcr.features

    def get_audiences(
        self,
    ) -> List[
        Union[
            RuleBasedAudienceDefinition
            | LookalikeAudienceDefinition
            | AdvertiserAudienceDefinition
        ]
    ]:
        """
        Get the audiences for the publisher.
        """
        if (
            not self.features.has_enable_lookalike_audiences()
            and not self.features.has_enable_remarketing()
            and not self.features.has_enable_rule_based_audiences()
        ):
            raise Exception(
                "Unable to retrieve audiences. Lookalike, remarketing or rule-based features must be enabled."
            )

        audiences_dict = self._request_audiences()
        audiences = []
        for a in audiences_dict["audiences"]:
            if a["kind"] == "rulebased":
                audiences.append(
                    RuleBasedAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            elif a["kind"] == "lookalike":
                audiences.append(
                    LookalikeAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            elif a["kind"] == "advertiser":
                audiences.append(
                    AdvertiserAudienceDefinition.from_dict(
                        a, audiences_dict["audiences"]
                    )
                )
            else:
                raise Exception(f'Audience kind "{a["kind"]}" unknown')

        return audiences

    def _request_audiences(self) -> Dict[str, Any]:
        """
        Retrieve the audiences data structure as provided by the backend.
        """
        computation = GetAudiencesForPublisherComputation(
            dcr_id=self.dcr_id, client=self.client, session=self.session
        )
        return computation.run_and_get_results()

    def estimate_audience_size(self, audience_name: str) -> int:
        """
        Estimate the audience size for the publisher.
        """
        audiences_json = self._request_audiences()
        payloads = get_parameter_payloads(
            audience_name=audience_name, audiences_json=audiences_json
        )

        if audience_depends_on_lookalike(
            audience_name=audience_name, audiences_json=audiences_json
        ):
            computation = EstimateAudienceSizeForPublisherLalComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
                lal_audience=payloads.lal,
            )
            result = computation.run_and_get_results()
            return result["audience_size"]
        else:
            computation = EstimateAudienceSizeForPublisherComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
            )
            result = computation.run_and_get_results()
            return result["audience_size"]

    def get_audience_user_list(self, audience_name: str) -> List[str]:
        """
        Get the audience user list for the publisher.
        """
        if (
            not self.features.has_enable_lookalike_audiences()
            and not self.features.has_enable_remarketing()
            and not self.features.has_enable_rule_based_audiences()
        ):
            raise Exception(
                "Unable to retrieve audience user list. Lookalike, remarketing or rule-based features must be enabled."
            )

        audiences_json = self._request_audiences()
        payloads = get_parameter_payloads(
            audience_name=audience_name, audiences_json=audiences_json
        )

        if audience_depends_on_lookalike(
            audience_name=audience_name, audiences_json=audiences_json
        ):
            computation = GetAudienceUserListForPublisherLalComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
                lal_audience=payloads.lal,
            )
            return computation.run_and_get_results()
        else:
            computation = GetAudienceUserListForPublisherComputation(
                dcr_id=self.dcr_id,
                client=self.client,
                session=self.session,
                generate_audience=payloads.generate,
            )
            return computation.run_and_get_results()

    def get_insights(self) -> Dict[str, Any]:
        """
        Get the insights.
        """
        if not self.features.has_enable_insights():
            raise Exception("This Audience Builder DCR does not support insights.")

        overlap_insights_computation = ComputeInsightsComputation(
            dcr_id=self.dcr_id, client=self.client, session=self.session
        )
        return overlap_insights_computation.run_and_get_results()

    def get_overlap_statistics(self) -> Dict[str, list[Any]]:
        """Get a report on how many users per audience were matched in the
        publisher's dataset. All numbers are rounded for privacy.

        **Returns**:
        A report with the following structure:

        ```
        {
            "overlap_statistics": [
                {
                    "audience_type": string,
                    "overlap_size": integer,
                    "andvertiser_size": integer
                }
            ]
        }
        ```

        If the DCR was configured to hide absolute audience counts, the
        values are omitted from the report.
        """
        return ComputeOverlapStatistics(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session,
        ).run_and_get_results()

    def get_validation_report(self) -> dict[str, Any]:
        """Get the validation reports for the advertiser datasets.

        Validation reports contain information about issues that were found
        in the provisioned data (e.g. values that don't match the expected schema
        or duplicated rows).

        **Returns**:
        A dictionary with keys "matching", "segments", "demographics". Each
        key points to the validation report of the respective dataset.
        """
        if not self.features.has_enable_drop_invalid_rows():
            raise Exception("This Audience Builder DCR does not validation of publisher datasets.")
        matching_report = GetMatchingValidationReport(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session
        ).run_and_get_results()
        segments_report = GetSegmentsValidationReport(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session
        ).run_and_get_results()
        demographics_report = GetDemographicsValidationReport(
            dcr_id=self.dcr_id,
            client=self.client,
            session=self.session
        ).run_and_get_results()
        return {
            "matching": matching_report,
            "segments": segments_report,
            "demographics": demographics_report,
        }

    def provision_from_data_lab(self, data_lab_id: str):
        """
        Provision the DataLab with the given ID to the Audience Builder DCR.

        **Parameters**:
        - `data_lab_id`: ID of the DataLab to provision to the Audience Builder DCR.
        """
        # First deprovision any existing datalabs before provisioning a new one.
        # This ensures that we don't get into issues with optional datasets not
        # being updated.
        self.deprovision_data_lab()

        # Check DataLab is validated
        data_lab = self.client.get_data_lab(data_lab_id)
        if not data_lab["isValidated"]:
            raise Exception("Cannot provision DataLab, not validated.")

        # Check compatibility
        hl_data_lab = data_lab["highLevelRepresentationAsString"]
        compatible = (
            ab_media_compiler.is_data_lab_compatible_with_ab_media_dcr_serialized(
                hl_data_lab, json.dumps(self.high_level_dcr)
            )
        )
        if not compatible:
            raise Exception("DataLab is incompatible with Audience Builder DCR")

        # Provision datasets
        data_lab_datasets = self._get_data_lab_datasets_dict(data_lab)

        # Provision all existing Data Lab datasets to the DCR.
        for dataset_type, dataset in data_lab_datasets.items():
            if not dataset:
                # Dataset was not provisioned to the Data Lab.
                continue
            manifest_hash = dataset["manifestHash"]
            encryption_key = self.client.get_dataset_key(manifest_hash)
            if dataset_type == "MATCHING_DATA":
                request_key = "publishMatchingDataset"
            elif dataset_type == "SEGMENTS_DATA":
                request_key = "publishSegmentsDataset"
            elif dataset_type == "DEMOGRAPHICS_DATA":
                request_key = "publishDemographicsDataset"
            elif dataset_type == "EMBEDDINGS_DATA":
                request_key = "publishEmbeddingsDataset"
            else:
                raise Exception(
                    f"Failed to provision Data Lab. Dataset type '{dataset_type}' unknown."
                )
            self._send_publish_dataset_request(
                request_key, manifest_hash, encryption_key
            )

        # Update DB.
        provisioned_data_lab_id = self.client._provision_data_lab_to_ab_dcr(
            self.dcr_id, data_lab_id
        )["id"]
        if provisioned_data_lab_id != data_lab_id:
            raise Exception(
                f"Incorrectly provisioned DataLab {provisioned_data_lab_id}"
            )

    def _get_data_lab_datasets_dict(self, data_lab: compiler.DataLab):
        datasets_dict = {}
        for dataset in data_lab["datasets"]:
            datasets_dict[dataset["name"]] = dataset["dataset"]
        return datasets_dict

    def _send_publish_dataset_request(
        self, request_key: str, manifest_hash: str, encryption_key: Key
    ):
        request = AbMediaRequest.model_validate(
            {
                request_key: {
                    "dataRoomIdHex": self.dcr_id,
                    "datasetHashHex": manifest_hash,
                    "encryptionKeyHex": encryption_key.material.hex(),
                    "scopeIdHex": self.client._ensure_dcr_data_scope(self.dcr_id),
                },
            }
        )
        response = Request.send(request, self.session)
        if request_key not in response.model_dump_json():
            raise Exception(f'Failed to publish "{request_key}"')

    # This function should not throw on repeated calls. This allows it to be
    # called when provisioning also (where there may not yet be any data provisioned).
    def deprovision_data_lab(self):
        """
        Deprovision a DataLab from the Audience Builder DCR.
        """
        # Get a list of published datasets
        request = AbMediaRequest.model_validate(
            {
                "retrievePublishedDatasets": {
                    "dataRoomIdHex": self.dcr_id,
                },
            }
        )
        response = Request.send(request, self.session)
        datasets = response.model_dump()["retrievePublishedDatasets"]
        if datasets["matchingDatasetHashHex"]:
            request_key = "unpublishMatchingDataset"
        elif datasets["segmentsDatasetHashHex"]:
            request_key = "unpublishSegmentsDataset"
        elif datasets["demographicsDatasetHashHex"]:
            request_key = "unpublishDemographicsDataset"
        elif datasets["embeddingsDatasetHashHex"]:
            request_key = "unpublishEmbeddingsDataset"
        else:
            request_key = None

        if request_key:
            self._send_unpublish_dataset_request(request_key)
            # Update DB.
            data_lab = self.client._deprovision_data_lab_from_ab_dcr(self.dcr_id)
            if data_lab:
                raise Exception("DataLab should have been deprovisioned")

    def _send_unpublish_dataset_request(self, request_key: str):
        request = AbMediaRequest.model_validate(
            {
                request_key: {
                    "dataRoomIdHex": self.dcr_id,
                },
            }
        )
        response = Request.send(request, self.session)
        if request_key not in response.model_dump_json():
            raise Exception(f"Failed to unpublish {request_key}")



import hashlib
import json
import os
from base64 import b64decode, b64encode, urlsafe_b64decode, urlsafe_b64encode
from concurrent import futures
from threading import BoundedSemaphore
from typing import BinaryIO, Dict, List, Optional, Tuple, cast, Generator
import io
import requests

from decentriq_dcr_compiler import compiler
from decentriq_dcr_compiler import ab_media as ab_media_compiler
from decentriq_dcr_compiler.schemas import DataScienceDataRoom
from decentriq_dcr_compiler.schemas import AbMediaDcr as AbMediaDcrSchema
from decentriq_dcr_compiler.schemas.secret_store_entry_state import SecretStoreEntryState as SecretStoreEntryStateSchema, v0 as v0_schema
from decentriq_dcr_compiler.types.secret_store_entry_state import v0

from .analytics import AnalyticsDcr, AnalyticsDcrDefinition
from .api import Api, Endpoints, NotFoundError, retry
from .attestation import enclave_specifications
from .authentication import Auth, generate_key, generate_self_signed_certificate
from .config import (
    _DECENTRIQ_UNSAFE_DISABLE_KNOWN_ROOT_CA_CHECK,
    DECENTRIQ_CLIENT_ID,
    DECENTRIQ_HOST,
    DECENTRIQ_PORT,
    DECENTRIQ_USE_TLS,
    DECENTRIQ_MRSIGNER_DRIVER_ATTESTATION_SPECIFICATION,
    DECENTRIQ_DISABLE_DATASET_ACL_ORGANIZATION_CHECK,
)
from .connection import Connection
from .endorsement import Endorser
from .graphql import GqlClient
from .ab_media import AbMediaDcr, AbMediaDcrDefinition
from .proto import AttestationSpecification, AuthenticationMethod, CreateDcrKind
from .proto import DataRoom as ProtoDataRoom
from .proto import PkiPolicy, parse_length_delimited, serialize_length_delimited
from .session import LATEST_WORKER_PROTOCOL_VERSION, Session
from .archv2 import Secret, SessionV2, ClientV2
from .storage import Chunker, Key, StorageCipher, create_encrypted_chunk, decrypt_chunk
from .types import (
    JSONType,
    CreateMediaComputeJobInput,
    DataLabDefinition,
    DataLabListFilter,
    DataRoom,
    DataRoomDescription,
    DataRoomKind,
    DatasetDescription,
    DatasetUsage,
    EnclaveSpecification,
    MediaComputeJob,
    MediaComputeJobFilterInput,
    OrganizationUser,
)
from .logger import logger

class SecretStoreOptions:
    def __init__(
        self,
        *,
        store_encryption_key: bool = True,
        encryption_key_acl: Optional[JSONType] = None,
        encryption_key_acl_version: int = 0,
    ):
        self.store_encryption_key = store_encryption_key
        if not store_encryption_key and encryption_key_acl is not None:
            raise ValueError("Encryption key ACL can only be set when storing the encryption key")
        self.encryption_key_acl = encryption_key_acl
        self.encryption_key_acl_version = encryption_key_acl_version

    def validate_users(self, client_user: str, organization_user: List[OrganizationUser]):
        # Validate the encryption key ACL
        if self.encryption_key_acl is not None:
            if self.encryption_key_acl_version == 0:
                # Check it's in valid format
                v0_schema.SecretStoreEntryAcl.model_validate(self.encryption_key_acl)
                validated_encryption_key_acl = cast(v0.SecretStoreEntryAcl, self.encryption_key_acl)
                if validated_encryption_key_acl["type"] == "UsersList":
                    # Check it contains at least one owner
                    if not any(user["role"] == "Owner" for user in validated_encryption_key_acl["users"]):
                        raise ValueError("Encryption key ACL must contain at least one owner")
                    # Check:
                    # - it contains only users from the organization

                    # Allow users to share datasets with other orgs (used for service users in the DMP)
                    if DECENTRIQ_DISABLE_DATASET_ACL_ORGANIZATION_CHECK:
                        return

                    organization_user_emails = set(user["email"] for user in organization_user)
                    for user in validated_encryption_key_acl["users"]:
                        if user["id"] == client_user:
                            continue
                        if user["id"] not in organization_user_emails:
                            raise ValueError("Encryption key ACL must contain only users from your organization")
                else:
                    raise ValueError(f"Unsupported encryption key ACL type {validated_encryption_key_acl['type']}")
            else:
                raise ValueError(f"Unsupported encryption key ACL version {self.encryption_key_acl_version}")

class Client:
    """
    A `Client` object allows you to upload datasets and to create `Session` objects that
    can communicate with enclaves and perform essential operations such as publishing
    data rooms and execute computations and retrieve results.

    Objects of this class can be used to create and run data rooms, as well as to securely
    upload data and retrieve computation results.

    Objects of this class should be created using the `create_client` function.
    """

    _api: Api
    _graphql: GqlClient
    _connections: Dict[str, Connection]
    _mrsigner_driver_spec: AttestationSpecification
    _cached_session_v2: Optional[SessionV2] = None
    client_v2: ClientV2

    def __init__(
        self,
        user_email: str,
        enclave_api_token: str,
        api: Api,
        graphql: GqlClient,
        request_timeout: Optional[int] = None,
        unsafe_disable_known_root_ca_check: bool = False,
        custom_mrsigner_driver_spec: Optional[AttestationSpecification] = None,
    ):
        """
        Create a client instance.

        Rather than creating `Client` instances directly using this constructor,
        use the function `create_client`.
        """
        self.user_email = user_email
        self.enclave_api_token = enclave_api_token
        self._api = api
        self._graphql = graphql
        self.request_timeout = request_timeout
        self.unsafe_disable_known_root_ca_check = unsafe_disable_known_root_ca_check
        self._connections = dict()
        if custom_mrsigner_driver_spec is not None:
            self._mrsigner_driver_spec = custom_mrsigner_driver_spec
        else:
            if DECENTRIQ_MRSIGNER_DRIVER_ATTESTATION_SPECIFICATION is not None:
                mrsigner_decoded_spec = AttestationSpecification()
                parse_length_delimited(
                    b64decode(DECENTRIQ_MRSIGNER_DRIVER_ATTESTATION_SPECIFICATION),
                    mrsigner_decoded_spec,
                )
                self._mrsigner_driver_spec = mrsigner_decoded_spec
            else:
                self._mrsigner_driver_spec = enclave_specifications.specifications["decentriq.driver:mrsigner"]["proto"]
        self.client_v2 = ClientV2(self._api)

    def check_enclave_availability(self, specs: Dict[str, EnclaveSpecification]):
        """
        Check whether the selected enclaves are deployed at this moment.
        If one of the enclaves is not deployed, an exception will be raised.
        """
        available_specs = [
            spec["proto"].SerializeToString()
            for spec in self._get_enclave_specifications()
        ]
        for spec in specs.values():
            if spec["proto"].SerializeToString() not in available_specs:
                raise Exception(
                    "No available enclave deployed for attestation spec '{name}' (version {version})".format(
                        name=spec["name"], version=spec["version"]
                    )
                )

    def _get_enclave_specifications(self) -> List[EnclaveSpecification]:
        data = self._graphql.post(
            """
            {
                attestationSpecsV2 {
                    name
                    version
                    spec
                }
            }
            """
        )
        enclave_specs = []
        for spec_json in data["attestationSpecsV2"]:
            attestation_specification = AttestationSpecification()
            spec_length_delimited = b64decode(spec_json["spec"])
            parse_length_delimited(spec_length_delimited, attestation_specification)
            enclave_spec = EnclaveSpecification(
                name=spec_json["name"],
                version=spec_json["version"],
                proto=attestation_specification,
                decoder=None,
                workerProtocols=[LATEST_WORKER_PROTOCOL_VERSION],
                clientProtocols=None,
            )
            enclave_specs.append(enclave_spec)
        return enclave_specs

    def create_session_from_data_room_description(
        self,
        data_room_description: DataRoomDescription,
        specs: Optional[List[EnclaveSpecification]] = None,
    ) -> Session:
        """
        Create a session for interacting with a DCR of the given data room description.
        """
        driver_attestation_hash = data_room_description["driverAttestationHash"]
        driver_enclave_spec = dict()
        specs = specs if specs else enclave_specifications.all()
        for spec in specs:
            attestation_hash = hashlib.sha256(
                serialize_length_delimited(spec["proto"])
            ).hexdigest()
            if attestation_hash == driver_attestation_hash:
                driver_enclave_spec = {"decentriq.driver": spec}
                break
        if not driver_enclave_spec:
            raise Exception(
                f"Driver enclave specification with attestation hash {driver_attestation_hash} not found"
            )
        auth, _ = self.create_auth_using_decentriq_pki(driver_enclave_spec)
        return self.create_session(auth, driver_enclave_spec)

    def create_session_v2(
        self,
    ) -> SessionV2:
        """
        Creates a new `decentriq_platform.session.SessionV2` instance to communicate
        with a driver enclave.
        """
        if self._cached_session_v2 is None:
            attestation_proto = self._mrsigner_driver_spec
            attestation_specification_hash = hashlib.sha256(
                serialize_length_delimited(attestation_proto)
            ).hexdigest()
            connection = self._connections.get(attestation_specification_hash)
            if connection is None:
                connection = Connection(
                    attestation_proto,
                    self._api,
                    self._graphql,
                    self.unsafe_disable_known_root_ca_check,
                )
                self._connections[attestation_specification_hash] = connection
            session = SessionV2(
                self,
                connection,
            )
            self._cached_session_v2 = session
        return self._cached_session_v2

    def create_session(
        self,
        auth: Auth,
        enclaves: Dict[str, EnclaveSpecification],
    ) -> Session:
        """
        Creates a new `decentriq_platform.session.Session` instance to communicate
        with a driver enclave.
        The passed set of enclave specifications must include a specification for
        a driver enclave.

        Messages sent through this session will be authenticated
        with the given authentication object.
        """
        if "decentriq.driver" not in enclaves:
            raise Exception(
                "Unable to find a specification for the driver enclave"
                + f" named 'decentriq.driver', you can get these specifications"
                + " from the main package."
            )
        driver_spec = enclaves["decentriq.driver"]
        if (
            "clientProtocols" not in driver_spec
            or driver_spec["clientProtocols"] is None
        ):
            raise Exception("Missing client supported protocol versions")
        attestation_proto = driver_spec["proto"]
        client_protocols = driver_spec["clientProtocols"]
        attestation_specification_hash = hashlib.sha256(
            serialize_length_delimited(attestation_proto)
        ).hexdigest()
        connection = self._connections.get(attestation_specification_hash)
        if connection is None:
            connection = Connection(
                attestation_proto,
                self._api,
                self._graphql,
                self.unsafe_disable_known_root_ca_check,
            )
            self._connections[attestation_specification_hash] = connection
        session = Session(
            self,
            connection,
            client_protocols,
            auth=auth,
        )

        return session

    def _ensure_dataset_scope(
        self,
        manifest_hash: Optional[str] = None,
    ) -> str:
        payload = {
            "manifestHash": manifest_hash,
        }
        data = self._graphql.post(
            """
            mutation GetOrCreateDatasetScope($input: CreateDatasetScopeInput!) {
                scope {
                    getOrCreateDatasetScope(input: $input) {
                        record {
                            id
                        }
                    }

                }
            }
            """,
            {"input": payload},
        )
        scope = data["scope"]["getOrCreateDatasetScope"]["record"]
        return scope["id"]

    def _ensure_dcr_data_scope(
        self,
        data_room_hash: str,
    ) -> str:
        data = self._graphql.post(
            """
            mutation GetOrCreateDcrDataScope($input: CreateDcrDataScopeInput!) {
                scope {
                    getOrCreateDcrDataScope(input: $input) {
                        record {
                            id
                        }
                    }

                }
            }
            """,
            {
                "input": {
                    "dataRoomHash": data_room_hash,
                }
            },
        )
        scope = data["scope"]["getOrCreateDcrDataScope"]["record"]
        return scope["id"]

    def _set_datalab_matching_dataset(
        self,
        data_lab_id: str,
        manifest_hash: Optional[str],
    ) -> str:
        """
        Store the matching dataset manifest hash in the database.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabUsersDataset($input: SetDataLabDatasetInput!) {
                dataLab {
                    setUsersDataset(input: $input) {
                        record {
                            id
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                        }
                    }

                }
            }
            """,
            {"input": {"id": data_lab_id, "manifestHash": manifest_hash}},
        )
        return data["dataLab"]["setUsersDataset"]["record"]["id"]

    def _set_datalab_segments_dataset(
        self,
        data_lab_id: str,
        manifest_hash: Optional[str],
    ) -> str:
        """
        Store the segments dataset manifest hash in the database.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabSegmentsDataset($input: SetDataLabDatasetInput!) {
                dataLab {
                    setSegmentsDataset(input: $input) {
                        record {
                            id
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                        }
                    }

                }
            }
            """,
            {"input": {"id": data_lab_id, "manifestHash": manifest_hash}},
        )
        return data["dataLab"]["setSegmentsDataset"]["record"]["id"]

    def _set_datalab_demographics_dataset(
        self,
        data_lab_id: str,
        manifest_hash: Optional[str],
    ) -> str:
        """
        Store the demographics dataset manifest hash in the database.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabDemographicsDataset($input: SetDataLabDatasetInput!) {
                dataLab {
                    setDemographicsDataset(input: $input) {
                        record {
                            id
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                        }
                    }

                }
            }
            """,
            {"input": {"id": data_lab_id, "manifestHash": manifest_hash}},
        )
        return data["dataLab"]["setDemographicsDataset"]["record"]["id"]

    def _set_datalab_embeddings_dataset(
        self,
        data_lab_id: str,
        manifest_hash: Optional[str],
    ) -> str:
        """
        Store the embeddings dataset manifest hash in the database.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabEmbeddingsDataset($input: SetDataLabDatasetInput!) {
                dataLab {
                    setEmbeddingsDataset(input: $input) {
                        record {
                            id
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                        }
                    }

                }
            }
            """,
            {"input": {"id": data_lab_id, "manifestHash": manifest_hash}},
        )
        return data["dataLab"]["setEmbeddingsDataset"]["record"]["id"]

    def _set_datalab_job_ids(
        self,
        data_lab_id: str,
        validation_compute_job_id: str,
        statistics_compute_job_id: str,
        jobs_driver_attestation_hash: str,
    ) -> str:
        """
        Store the job IDs associated with the DataLab.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabJobIds($input: SetDataLabJobIdsInput!) {
                dataLab {
                    setJobIds(input: $input) {
                        record {
                            id
                        }
                    }

                }
            }
            """,
            {
                "input": {
                    "id": data_lab_id,
                    "validationComputeJobId": validation_compute_job_id,
                    "statisticsComputeJobId": statistics_compute_job_id,
                    "jobsDriverAttestationHash": jobs_driver_attestation_hash,
                }
            },
        )
        return data["dataLab"]["setJobIds"]["record"]["id"]

    def _set_datalab_statistics(
        self,
        data_lab_id: str,
        statistics: str,
    ) -> str:
        """
        Store the DataLab statistics in the database.
        """
        data = self._graphql.post(
            """
            mutation SetDataLabStatistics($input: SetDataLabStatisticsInput!) {
                dataLab {
                    setStatistics(input: $input) {
                        record {
                            id
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "id": data_lab_id,
                    "statistics": json.loads(statistics),
                }
            },
        )
        return data["dataLab"]["setStatistics"]["record"]["id"]

    def _iter_response_encrypted_chunks(self, response):
        expecting_header = True
        buffer = bytearray()
        chunk_size = None
        for chunk in response.iter_content(chunk_size=8 * 1024 * 1024 + 4096):
            buffer += chunk
            while True:
                if expecting_header:
                    splits = buffer.split(b'\n', 1)
                    if len(splits) == 1:
                        break
                    if len(splits) != 2:
                        raise Exception(f"Error during stream decoding: Expected header and data to be separated by a newline")
                    json_header = json.loads(splits[0])
                    if "V0" not in json_header:
                        raise Exception(f"Error during stream decoding: Expected V0 header")
                    if "Failure" in json_header["V0"]:
                        raise Exception(f"Error during stream decoding: {json_header['V0']['Failure']['error']}")
                    if "Next" not in json_header["V0"]:
                        raise Exception(f"Error during stream decoding: Expected Next header")
                    chunk_size = json_header["V0"]["Next"]['size']
                    if chunk_size > 8 * 1024 * 1024 + 4096:
                        raise Exception(f"Chunk size {chunk_size} is too large")
                    expecting_header = False
                    buffer = splits[1]

                if len(buffer) < chunk_size:
                    break
                else:
                    yield memoryview(buffer)[:chunk_size]
                    buffer = buffer[chunk_size:]
                    chunk_size = None
                    expecting_header = True

        if len(buffer) > 0:
            raise Exception(f"Error during stream decoding: unexpected trailing data")

    def _iter_response_decrypted(self, response: requests.Response, key: Key, chunk_hashes: List[str]) -> Generator[bytes, None, None]:
        for chunk_hash, encrypted_chunk in zip(chunk_hashes, self._iter_response_encrypted_chunks(response)):
            yield decrypt_chunk(key, chunk_hash, encrypted_chunk)

    def _stream_response_decrypted(self, response: requests.Response, key: Key, chunk_hashes: List[str]) -> io.RawIOBase:
        return GeneratorStream(self._iter_response_decrypted(response, key, chunk_hashes))

    def download_dataset(self, manifest_hash: str, key: Key) -> io.RawIOBase:
        # First get the manifest
        response = self._api.post(
            Endpoints.DATASET_CHUNKS_STREAM.replace(":manifest_hash", manifest_hash),
            headers={"Content-type": "application/octet-stream"},
            req_body=manifest_hash,
        )
        chunk_hashes = json.loads(self._stream_response_decrypted(response, key, [manifest_hash]).read())

        # Now download the chunks
        response = self._api.post(
            Endpoints.DATASET_CHUNKS_STREAM.replace(":manifest_hash", manifest_hash),
            headers={"Content-type": "application/octet-stream"},
            req_body="\n".join(chunk_hashes),
        )

        return self._stream_response_decrypted(response, key, chunk_hashes)

    def download_job_result(self, job_id: str, manifest_hash: str, task_result_hash: str, key: Key) -> io.RawIOBase:
        endpoint = Endpoints.JOB_RESULT_CHUNKS_STREAM.replace(":job_id", job_id).replace(
            ":task_result_hash", task_result_hash
        )

        # First get the manifest
        response = self._api.post(
            endpoint,
            headers={"Content-type": "application/octet-stream"},
            req_body=manifest_hash,
        )
        chunk_hashes = json.loads(
            self._stream_response_decrypted(response, key, [manifest_hash]).read()
        )

        # Now download the chunks
        response = self._api.post(
            endpoint,
            headers={"Content-type": "application/octet-stream"},
            req_body="\n".join(chunk_hashes),
        )

        return self._stream_response_decrypted(response, key, chunk_hashes)

    def upload_dataset(
        self,
        data: BinaryIO,
        key: Key,
        file_name: str,
        /,
        *,
        description: str = "",
        chunk_size: int = 8 * 1024**2,
        parallel_uploads: int = 8,
        usage: DatasetUsage = DatasetUsage.PUBLISHED,
        secret_store_options: Optional[SecretStoreOptions] = None,
        is_accessory: bool = False
    ) -> str:
        """
        Uploads `data` as a file usable by enclaves and returns the
        corresponding manifest hash.

        **Parameters**:
        - `data`: The data to upload as a buffered stream.
            Such an object can be obtained by wrapping a binary string in a `io.BytesIO()`
            object or, if reading from a file, by using `with open(path, "rb") as file`.
        - `key`: Encryption key used to encrypt the file.
        - `file_name`: Name of the file.
        - `description`: An optional file description.
        - `chunk_size`: Size of the chunks into which the stream is split in bytes.
        - `parallel_uploads`: Whether to upload chunks in parallel.
        - `usage`: The usage of the dataset.
        - `secret_store_options`: Options for the secret store.
            It can be used to specify if the encryption key should be stored in the secret store
            and can also be used to provide a custom ACL for the encryption key.
        - `is_accessory`: Whether this dataset should be hidden from the datasets page.
        """
        if secret_store_options is None:
            secret_store_options = SecretStoreOptions()
        else:
            secret_store_options.validate_users(self.user_email, self._get_my_organization_users())
        uploader = BoundedExecutor(
            bound=parallel_uploads * 2, max_workers=parallel_uploads
        )
        # create and upload chunks
        chunker = Chunker(data, chunk_size=chunk_size)
        chunk_hashes: List[str] = []
        chunk_content_sizes: List[int] = []
        chunk_uploads_futures = []
        upload_id = self._create_upload()

        for chunk_hash, chunk_data, chunk_content_size in chunker:
            chunk_uploads_futures.append(
                uploader.submit(
                    self._encrypt_and_upload_chunk,
                    chunk_hash,
                    chunk_data,
                    key.material,
                    upload_id,
                )
            )
            chunk_hashes.append(chunk_hash.hex())
            chunk_content_sizes.append(chunk_content_size)

        # check chunks uploads were successful
        completed, pending = futures.wait(
            chunk_uploads_futures, None, futures.FIRST_EXCEPTION
        )
        # re-raise exception
        for future in completed:
            future.result()
        uploader.shutdown(wait=False)

        # create manifest and upload
        manifest_hash_bytes, manifest_encrypted = create_encrypted_chunk(
            key.material,
            os.urandom(16),
            json.dumps(chunk_hashes).encode("utf-8"),
            content_size=chunker.content_size,
            chunk_content_sizes=chunk_content_sizes,
        )
        scope_id = self._ensure_dataset_scope(
            manifest_hash_bytes.hex(),
        )
        manifest_hash = self._finalize_upload(
            scope_id=scope_id,
            upload_id=upload_id,
            name=file_name,
            manifest_hash_bytes=manifest_hash_bytes,
            manifest_encrypted=manifest_encrypted,
            chunks=chunk_hashes,
            description=description,
            usage=usage,
            size=sum(chunk_content_sizes),
            is_accessory=is_accessory,
        )

        if secret_store_options.store_encryption_key:
            session_v2 = self.create_session_v2()

            if secret_store_options.encryption_key_acl_version == 0:
                acl = secret_store_options.encryption_key_acl if secret_store_options.encryption_key_acl is not None else {
                    "type": "UsersList",
                    "users": [
                        {
                            "id": self.user_email,
                            "role": "Owner",
                        },
                    ],
                }
                encryption_key_secret = Secret(secret=key.material, state=SecretStoreEntryStateSchema.model_validate(
                    {
                        "version": "V0",
                        "acl": acl,
                        "type": "DatasetKey",
                        "manifest_hash": manifest_hash,
                    }
                ))
            else:
                raise ValueError(f"Unsupported encryption key ACL version {secret_store_options.encryption_key_acl_version}")
            session_v2.create_secret(encryption_key_secret)

        return manifest_hash

    def get_dataset_key(self, manifest_hash: str) -> Key:
        dataset_encryption_key_secret_id = self.get_dataset_encryption_key_secret_id(manifest_hash)
        if dataset_encryption_key_secret_id is None:
            raise Exception(f"Dataset with manifest hash {manifest_hash} has no associated secret")
        session_v2 = self.create_session_v2()
        dataset_encryption_key_secret, _ = session_v2.get_secret(dataset_encryption_key_secret_id)
        dataset_key = dataset_encryption_key_secret.secret
        return Key(dataset_key)

    def get_dataset_encryption_key_secret_id(self, manifest_hash: str) -> Optional[str]:
        dataset = self.get_dataset(manifest_hash)
        if dataset is None:
            raise Exception(f"Dataset with manifest hash {manifest_hash} not found")
        return dataset["encryptionKeySecretId"]

    def _encrypt_and_upload_chunk(
        self, chunk_hash: bytes, chunk_data: bytes, key: bytes, upload_id: str
    ):
        cipher = StorageCipher(key)
        chunk_data_encrypted = cipher.encrypt(chunk_data)
        self._upload_chunk(chunk_hash, chunk_data_encrypted, upload_id)

    def _create_upload(self) -> str:
        """
        Create an upload record for the user identified by the used
        API token and return its id.
        """
        data = self._graphql.post(
            """
            mutation CreateUpload() {
                upload {
                    create {
                        record {
                            id
                        }
                    }
                }
            }
            """,
        )
        upload_id = data["upload"]["create"]["record"]["id"]
        return upload_id

    def _upload_chunk(
        self, chunk_hash: bytes, chunk_data_encrypted: bytes, upload_id: str
    ):
        url = Endpoints.USER_UPLOAD_CHUNKS.replace(":uploadId", upload_id).replace(
            ":chunkHash", chunk_hash.hex()
        )
        try:
            self._api.put(
                url,
                chunk_data_encrypted,
                {"Content-type": "application/octet-stream"},
                retry=retry,
            )
        except Exception as e:
            logger.error(e)
            raise e

    def _delete_user_upload(self, upload_id: str):
        self._graphql.post(
            """
            mutation DeleteUpload($id: Id!) {
                upload {
                    delete(id: $id)
                }
            }
            """,
            {"id": upload_id},
        )

    def _finalize_upload(
        self,
        scope_id: str,
        upload_id: str,
        name: str,
        manifest_hash_bytes: bytes,
        manifest_encrypted: bytes,
        chunks: List[str],
        size: int,
        description: Optional[str] = None,
        usage: Optional[DatasetUsage] = None,
        is_accessory: bool = False,
    ) -> str:
        data = self._graphql.post(
            """
            mutation FinalizeUpload($input: CreateDatasetForUploadInput!) {
                upload {
                    finalizeUploadAndCreateDataset(input: $input) {
                        record {
                            id
                            manifestHash
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "uploadId": upload_id,
                    "manifest": b64encode(manifest_encrypted).decode("ascii"),
                    "manifestHash": manifest_hash_bytes.hex(),
                    "name": name,
                    "description": description,
                    "usage": usage,
                    "chunkHashes": chunks,
                    "scopeId": scope_id,
                    "size": size,
                    "isAccessory": is_accessory,
                }
            },
            retry=retry,
        )

        dataset = data["upload"]["finalizeUploadAndCreateDataset"]["record"]
        manifest_hash = dataset["manifestHash"]

        return manifest_hash

    def get_dataset(self, manifest_hash: str) -> Optional[DatasetDescription]:
        """
        Returns information about a user dataset given a dataset id.
        """
        try:
            data = self._graphql.post(
                """
                query GetDataset($manifestHash: HexString!)
                {
                    datasetByManifestHash(manifestHash: $manifestHash) {
                        id
                        name
                        manifestHash
                        description
                        createdAt
                        encryptionKeySecretId
                        metadataSecretId
                    }
                }
                """,
                {"manifestHash": manifest_hash},
            )
            return data["datasetByManifestHash"]
        except NotFoundError:
            return None

    def get_available_datasets(self) -> List[DatasetDescription]:
        """
        Returns the a list of datasets that the current user uploaded,
        regardless of whether they have already been connected to a
        data room or not.
        """
        data = self._graphql.post(
            """
            {
                myself {
                    datasets {
                        nodes {
                            id
                            name
                            manifestHash
                            statistics
                            size
                            description
                            createdAt
                            usage
                        }
                    }
                }
            }
            """
        )
        return data["myself"]["datasets"]["nodes"]

    def delete_dataset(self, manifest_hash: str, force: bool = False):
        """
        Deletes the dataset with the given id from the Decentriq platform.

        In case the dataset is still published to one or more data rooms,
        an exception will be thrown and the dataset will need to be
        unpublished manually from the respective data rooms using
        `Session.remove_published_dataset`.
        This behavior can be overridden by using the `force` flag.
        Note, however, that this might put some data rooms in a broken
        state as they might try to read data that does not exist anymore.
        """
        data_rooms_ids_with_dataset = self._get_data_room_ids_with_published_dataset(
            manifest_hash
        )
        if data_rooms_ids_with_dataset:
            id_list = "\n".join(
                [f"- {dcr_id}" for dcr_id in data_rooms_ids_with_dataset]
            )
            if force:
                logger.warning(
                    "This dataset is published to the following data rooms."
                    " These data rooms might be in a broken state now:"
                    f"\n{id_list}"
                )
            else:
                raise Exception(
                    "This dataset is published to the following data rooms"
                    " and needs to be unpublished before it can be deleted!"
                    f"\n{id_list}"
                )
        self._graphql.post(
            """
            mutation DeleteDataset($manifestHash: HexString!) {
                dataset {
                    deleteByManifestHash(manifestHash: $manifestHash)
                }
            }
            """,
            {
                "manifestHash": manifest_hash,
            },
        )

    @property
    def decentriq_ca_root_certificate(self) -> bytes:
        """
        Returns the root certificate used by the Decentriq identity provider.
        Note that when using this certificate in any authentication scheme,
        you trust Decentriq as an identity provider!
        """
        data = self._graphql.post(
            """
            {
                certificateAuthority {
                    rootCertificate
                }
            }
        """
        )
        certificate = data["certificateAuthority"]["rootCertificate"].encode("utf-8")
        return certificate

    @property
    def decentriq_pki_authentication(self) -> AuthenticationMethod:
        """
        The authentication method that uses the Decentriq root certificate to authenticate
        users.

        This method should be specified when building a data room in case you want to interact
        with the that data room either via the web interface or with sessions created using
        `create_auth_using_decentriq_pki`.
        Note that when using this authentication method you trust Decentriq as an identity provider!

        You can also create an `AuthenticationMethod` object directly and supply your own root certificate,
        with which to authenticate users connecting to your data room.
        In this case you will also need to issue corresponding user certificates and create your
        own custom `decentriq_platform.authentication.Auth` objects.
        """
        root_pki = self.decentriq_ca_root_certificate
        return AuthenticationMethod(dqPki=PkiPolicy(rootCertificatePem=root_pki))

    def create_auth_using_decentriq_pki(
        self, enclaves: Dict[str, EnclaveSpecification]
    ) -> Tuple[Auth, Endorser]:
        auth = self.create_auth()
        endorser = Endorser(auth, self, enclaves)
        dq_pki = endorser.decentriq_pki_endorsement()
        auth.attach_endorsement(decentriq_pki=dq_pki)
        return auth, endorser

    def create_auth(self) -> Auth:
        """
        Creates a `decentriq_platform.authentication.Auth` object which can be attached
        to `decentriq_platform.session.Session`.
        """
        keypair = generate_key()
        cert_chain_pem = generate_self_signed_certificate(self.user_email, keypair)
        auth = Auth(cert_chain_pem, keypair, self.user_email)
        return auth

    def get_data_room_descriptions(self, *, exclude_stopped_dcrs: bool = False) -> List[DataRoomDescription]:
        """
        Returns a list of data room descriptions that a user has created or
        participates in.

        Setting `exclude_stopped_dcrs` to `True` omits stopped data room descriptions
        from the returned list.
        """
        data = self._graphql.post(
            """
            {
                publishedDataRooms {
                    nodes {
                        id
                        title
                        driverAttestationHash
                        isStopped
                        createdAt
                        updatedAt
                        owner {
                            email
                        }
                        kind
                    }
                }
            }
            """
        )
        all_dcr_descriptions = [
            DataRoomDescription(**item) for item in data["publishedDataRooms"]["nodes"]
        ]
        if exclude_stopped_dcrs:
            active_dcr_descriptions = [
                d for d in all_dcr_descriptions if d["isStopped"] == False
            ]
            return active_dcr_descriptions
        else:
            return all_dcr_descriptions

    def get_data_room_description(
            self, data_room_hash, enclave_specs=None
    ) -> Optional[DataRoomDescription]:
        """
        Get a single data room description.
        """
        if enclave_specs is not None:
            driver_spec = enclave_specs["decentriq.driver"]
            attestation_proto = driver_spec["proto"]
            driver_attestation_hash = hashlib.sha256(
                serialize_length_delimited(attestation_proto)
            ).hexdigest()
        else:
            driver_attestation_hash = None
        return self._get_data_room_by_hash(data_room_hash, driver_attestation_hash)

    def _get_data_room_kind(
        self,
        data_room_id: str,
    ) -> DataRoomKind:
        """
        Get the kind of data room.
        """
        data = self._graphql.post(
            """
            query GetPublishedDataRoomType($dataRoomId: String!) {
                publishedDataRoom(id: $dataRoomId) {
                    kind
                }
            }
            """,
            {
                "dataRoomId": data_room_id,
            },
        )
        return data["publishedDataRoom"]["kind"]

    def _get_data_room_by_hash(
        self, data_room_hash: str, driver_attestation_hash: Optional[str] = None
    ) -> Optional[DataRoomDescription]:
        data = self._graphql.post(
            """
            query GetPublishedDataRoom($dataRoomHash: String!) {
                publishedDataRoom(id: $dataRoomHash) {
                    id
                    title
                    driverAttestationHash
                    isStopped
                    createdAt
                    updatedAt
                    owner {
                        email
                    }
                    kind
                }
            }
            """,
            {
                "dataRoomHash": data_room_hash,
            },
        )
        result = data.get("publishedDataRoom")
        if result is not None:
            dcr: DataRoomDescription = result
            if driver_attestation_hash is not None and dcr["driverAttestationHash"] != driver_attestation_hash:
                raise Exception(
                    f"Driver attestation hash for request dataroom doesn't match '{dcr['driverAttestationHash']}' != {driver_attestation_hash})"
                )
            return dcr
        else:
            return None

    def _get_user_certificate(self, email: str, csr_pem: str) -> str:
        data = self._graphql.post(
            """
            query getUserCertificate($input: UserCsrInput!) {
                certificateAuthority {
                    userCertificate(input: $input)
                }
            }
            """,
            {
                "input": {
                    "csrPem": csr_pem,
                    "email": email,
                }
            },
        )
        cert_chain_pem = data["certificateAuthority"]["userCertificate"]
        return cert_chain_pem

    def _get_data_room_ids_with_published_dataset(
        self, manifest_hash
    ) -> List[str]:
        data = self._graphql.post(
            """
                query GetDatasetPublications($manifestHash: HexString!) {
                    datasetByManifestHash(manifestHash: $manifestHash) {
                        publications {
                            nodes {
                                dataRoom {
                                    id
                                }
                            }
                        }
                    }
                }
                """,
            {
                "manifestHash": manifest_hash,
            },
        )
        publications = data["datasetByManifestHash"]["publications"]["nodes"]

        if publications:
            dcr_ids = list(set([publication["dataRoom"]["id"] for publication in publications]))
            return dcr_ids
        else:
            return []

    def _get_scope(self, scope_id: str):
        data = self._graphql.post(
            """
            query GetScope($scopeId: String!) {
                scope(id: $scopeId) {
                    id
                    organization {
                        id
                        name
                    }
                    owner {
                        id
                        email
                    }
                    scopeType
                    manifestHash
                    dataRoomHash
                    driverAttestationHash
                    createdAt
                }
            }
            """,
            {"scopeId": scope_id},
        )
        return data["scope"]

    def list_data_labs(
        self, filter: Optional[DataLabListFilter] = None
    ) -> List[DataLabDefinition]:
        """
        Return a list of DataLabs based on the `filter` criteria.

        **Parameters**:
        - `filter`: Criteria used to filter the list. Can be one of the following values:
            - NONE: Display all DataLabs.
            - VALIDATED: Display DataLabs that have been validated.
            - UNVALIDATED: Display DataLabs that have not been validated.
        """
        data = self._graphql.post(
            """
            query ListDataLabIds() {
                dataLabs {
                    nodes {
                        id
                        name
                        datasets {
                            name
                            dataset {
                                id
                                manifestHash
                                name
                            }
                        }
                        usersDataset {
                            id
                            manifestHash
                            name
                        }
                        segmentsDataset {
                            id
                            manifestHash
                            name
                        }
                        demographicsDataset {
                            id
                            manifestHash
                            name
                        }
                        embeddingsDataset  {
                            id
                            manifestHash
                            name
                        }
                        requireDemographicsDataset
                        requireEmbeddingsDataset
                        isValidated
                        numEmbeddings
                        matchingIdFormat
                        matchingIdHashingAlgorithm
                        validationComputeJobId
                        statisticsComputeJobId
                        jobsDriverAttestationHash
                        highLevelRepresentationAsString
                        createdAt
                        updatedAt
                    }
                }
            }
            """,
        )
        data_labs = data["dataLabs"]["nodes"]
        if filter is None:
            return [lab for lab in data_labs]
        elif filter == DataLabListFilter.VALIDATED:
            return [lab for lab in data_labs if lab["isValidated"] == True]
        elif filter == DataLabListFilter.UNVALIDATED:
            return [lab for lab in data_labs if lab["isValidated"] == False]
        else:
            raise Exception(f"Unknown DataLab filter {filter}")

    def get_data_lab(
        self,
        id: str,
    ) -> DataLabDefinition:
        """
        Return the DataLab with the given ID.

        **Parameters**:
        - `id`: ID of the DataLab to get.
        """
        data = self._graphql.post(
            """
            query GetDataLab($id: String!) {
                dataLab(id: $id) {
                    id
                    name
                    datasets {
                        name
                        dataset {
                            id
                            manifestHash
                            name
                        }
                    }
                    usersDataset {
                        id
                        manifestHash
                        name
                    }
                    segmentsDataset {
                        id
                        manifestHash
                        name
                    }
                    demographicsDataset {
                        id
                        manifestHash
                        name
                    }
                    embeddingsDataset {
                        id
                        manifestHash
                        name
                    }
                    statistics
                    requireSegmentsDataset
                    requireDemographicsDataset
                    requireEmbeddingsDataset
                    isValidated
                    numEmbeddings
                    matchingIdFormat
                    matchingIdHashingAlgorithm
                    validationComputeJobId
                    statisticsComputeJobId
                    jobsDriverAttestationHash
                    highLevelRepresentationAsString
                    forceSparkValidation
                    dropInvalidRows
                }
            }
            """,
            {"id": id},
        )
        return data["dataLab"]

    def _publish_data_lab_from_existing(
        self,
        data_lab: Dict[str, str],
    ) -> str:
        """
        Publish a DataLab from an existing high level representation.

        **Parameters**:
        - `data_lab`: DataLab high level representation
        """
        data = self._graphql.post(
            """
            mutation PublishDataLab($input: CreateDataLabFromExistingInput!) {
                dataLab {
                    createFromExisting(input: $input) {
                        record {
                            id
                        }
                    }
                }
            }
            """,
            {"input": {"dataLab": data_lab}},
        )
        return data["dataLab"]["createFromExisting"]["record"]["id"]

    def _get_enclave_spec_from_hash(self, hash: str) -> Optional[EnclaveSpecification]:
        available_specs = self._get_enclave_specifications()
        for spec in available_specs:
            hashed_attestation_spec = hashlib.sha256(
                serialize_length_delimited(spec["proto"])
            ).hexdigest()
            if hashed_attestation_spec == hash:
                return spec
        return None

    def _create_media_compute_job(
        self,
        input: CreateMediaComputeJobInput,
    ) -> MediaComputeJob:
        """
        Create a compute job for the Media DCR.
        """
        data = self._graphql.post(
            """
            mutation CreateMediaComputeJob($input: CreateMediaComputeJobInput!) {
                mediaComputeJob {
                    create(input: $input) {
                        record {
                            jobIdHex
                            publishedDataRoomId
                            computeNodeName
                            jobType
                            cacheKey
                            createdAt
                        }
                    }
                }
            }
            """,
            {"input": input},
        )
        return data["mediaComputeJob"]["create"]["record"]

    def _get_media_compute_job(
        self,
        input: MediaComputeJobFilterInput,
    ) -> MediaComputeJob:
        """
        Get a compute job for the Media DCR.
        """
        data = self._graphql.post(
            """
            query GetMediaComputeJob($input: MediaComputeJobFilterInput!) {
                mediaComputeJob (input: $input) {
                    jobIdHex
                    publishedDataRoomId
                    computeNodeName
                    jobType
                    cacheKey
                    createdAt
                }
            }
            """,
            {"input": input},
        )
        return data["mediaComputeJob"]

    def _provision_data_lab(
        self,
        data_room_id: str,
        data_lab_id: str,
    ) -> DataLabDefinition:
        """
        Provision a DataLab to a DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to provision to.
        - `data_lab_id`: ID of the DataLab to be provisioned.
        """
        data = self._graphql.post(
            """
            mutation ProvisionDataLab($input: ProvisionDataLabInput!) {
                dataLab {
                    provisionDataLab(input: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "dataRoomId": data_room_id,
                    "dataLabId": data_lab_id,
                }
            },
        )
        return data["dataLab"]["provisionDataLab"]["publishedDataLab"]

    def _deprovision_data_lab(self, data_room_id: str) -> DataLabDefinition:
        """
        Deprovision a DataLab from a DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to deprovision from.
        """
        data = self._graphql.post(
            """
            mutation DeprovisionDataLab($input: String!) {
                dataLab {
                    deprovisionDataLab(lookalikeMediaDcrId: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": data_room_id,
            },
        )
        return data["dataLab"]["deprovisionDataLab"]["publishedDataLab"]

    def retrieve_analytics_dcr(
        self,
        dcr_id,
        enclave_specs: Optional[List[EnclaveSpecification]] = None,
    ) -> AnalyticsDcr:
        """
        Retrieve an existing Analytics DCR.

        **Parameters**:
        - `dcr_id`: Data Clean Room ID.
        - `enclave_specs`: The enclave specifications that are considered
          to be trusted. If not specified, all enclave specifications known
          to this version of the SDK will be used.
        """
        return AnalyticsDcr._from_existing(
            dcr_id, client=self, enclave_specs=enclave_specs
        )

    def publish_analytics_dcr(
        self,
        dcr_definition: AnalyticsDcrDefinition,
        *,
        enclave_specs: Optional[Dict[str, EnclaveSpecification]] = None,
    ) -> AnalyticsDcr:
        """
        Publish an Analytics DCR.

        **Parameters**:
        - `dcr_definition`: Definition of the Analytics DCR.
        - `enclave_specs`: The enclave specifications that are considered
          to be trusted. If not specified, all enclave specifications known
          to this version of the SDK will be used.
        """
        hl = dcr_definition._get_high_level_representation()

        data_room = DataScienceDataRoom.model_validate(hl)
        compiled_data_room = compiler.compile_data_science_data_room(data_room)
        self.compile_context = compiled_data_room.compile_context

        low_level_data_room = ProtoDataRoom()
        parse_length_delimited(compiled_data_room.data_room, low_level_data_room)

        # Get a new session.
        # Determine which driver enclave spec (as given by the enclave_specs value)
        # to use. If this is not explicitly specified, try to check whether it was
        # already set on the builder that constructed the DCR definition.
        # If this is also not specified, simply use the latest specifications known to this SDK.
        specs = (
            enclave_specs
            or dcr_definition.enclave_specs
            or enclave_specifications.latest()
        )
        auth, _ = self.create_auth_using_decentriq_pki(specs)
        session = self.create_session(auth, specs)

        dcr_id = session.publish_data_room(
            low_level_data_room,
            kind=CreateDcrKind.DATASCIENCE,
            high_level_representation=compiled_data_room.datascience_data_room_encoded,
        )

        # Now that we've published the DCR the simplest way to construct
        # the DCR is using the `from_existing` method. This takes care
        # of correctly constructing all the node definitions.
        published_ds_dcr = AnalyticsDcr._from_existing(
            dcr_id=dcr_id, client=self, enclave_specs=list(specs.values())
        )

        data_room_description = self.get_data_room_description(dcr_id, specs)
        if not data_room_description:
            raise Exception(f"Failed to get data room description for DCR ID {dcr_id}")

        # Notify participants of DCR creation.
        data = self._graphql.post(
            """
            mutation NotifyParticipants($input: InviteParticipantsInput!) {
                publishedDataRoom {
                    inviteParticipants(input: $input) {
                        id
                    }
                }
            }
            """,
            {
                "input": {
                    "publishedDataRoomEnclaveId": dcr_id,
                    "publishedDataRoomDriverAttestationHash": data_room_description[
                        "driverAttestationHash"
                    ],
                    "dataRoomDescription": dcr_definition._get_description(),
                },
            },
        )

        return published_ds_dcr

    def retrieve_ab_media_dcr(
        self,
        dcr_id,
        enclave_specs: Optional[List[EnclaveSpecification]] = None,
    ) -> AbMediaDcr:
        """
        Retrieve an existing Audience Builder DCR.

        **Parameters**:
        - `dcr_id`: Data Clean Room ID.
        - `enclave_specs`: The enclave specifications that are considered
          to be trusted. If not specified, all enclave specifications known
          to this version of the SDK will be used.
        """
        return AbMediaDcr._from_existing(
            dcr_id, client=self, enclave_specs=enclave_specs
        )

    def publish_ab_media_dcr(
        self,
        dcr_definition: AbMediaDcrDefinition,
        *,
        enclave_specs: Optional[Dict[str, EnclaveSpecification]] = None,
    ) -> AbMediaDcr:
        """
        Publish an Audience Builder DCR.

        **Parameters**:
        - `dcr_definition`: Definition of the Audience Builder DCR.
        - `enclave_specs`: The enclave specifications that are considered
          to be trusted. If not specified, all enclave specifications known
          to this version of the SDK will be used.
        """
        dcr = AbMediaDcrSchema.model_validate_json(
            json.dumps(dcr_definition._high_level)
        )
        # Ensure we create the latest known version of the DCR.
        dcr_latest = ab_media_compiler.upgrade_ab_media_dcr_to_latest(dcr)

        compiled_serialized = ab_media_compiler.compile_ab_media_dcr(dcr_latest)
        low_level_dcr = ProtoDataRoom()
        parse_length_delimited(compiled_serialized, low_level_dcr)

        # Get a new session.
        # Determine which driver enclave spec (as given by the enclave_specs value)
        # to use. If this is not explicitly specified, try to check whether it was
        # already set on the builder that constructed the DCR definition.
        # If this is also not specified, simply use the latest specifications known to this SDK.
        specs = (
            enclave_specs
            or dcr_definition._enclave_specs
            or enclave_specifications.latest()
        )
        auth, _ = self.create_auth_using_decentriq_pki(specs)
        session = self.create_session(auth, specs)

        dcr_id = session.publish_data_room(
            low_level_dcr,
            kind=CreateDcrKind.AB_MEDIA,
            high_level_representation=dcr_latest.model_dump_json(
                by_alias=True
            ).encode(),
        )
        existing_dcr = AbMediaDcr._from_existing(
            dcr_id=dcr_id,
            client=self,
            enclave_specs=list(specs.values()),
        )
        return existing_dcr

    def _provision_data_lab_to_midcr(
        self,
        data_room_id: str,
        data_lab_id: str,
    ) -> DataLabDefinition:
        """
        Provision a DataLab to a Media DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to provision to.
        - `data_lab_id`: ID of the DataLab to be provisioned.
        """
        data = self._graphql.post(
            """
            mutation ProvisionDataLabToMediaInsightsDcr($input: ProvisionDataLabInput!) {
                dataLab {
                    provisionDataLabToMediaInsightsDcr(input: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "dataRoomId": data_room_id,
                    "dataLabId": data_lab_id,
                }
            },
        )
        return data["dataLab"]["provisionDataLabToMediaInsightsDcr"]["publishedDataLab"]

    def _deprovision_data_lab_from_midcr(self, data_room_id: str) -> DataLabDefinition:
        """
        Deprovision a DataLab from a Media DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to deprovision from.
        """
        data = self._graphql.post(
            """
            mutation DeprovisionDataLabFromMediaInsightsDcr($input: String!) {
                dataLab {
                    deprovisionDataLabFromMediaInsightsDcr(mediaInsightsDcrId: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": data_room_id,
            },
        )
        return data["dataLab"]["deprovisionDataLabFromMediaInsightsDcr"][
            "publishedDataLab"
        ]

    def _provision_data_lab_to_ab_dcr(
        self,
        data_room_id: str,
        data_lab_id: str,
    ) -> DataLabDefinition:
        """
        Provision a DataLab to an Audience Builder DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to provision to.
        - `data_lab_id`: ID of the DataLab to be provisioned.
        """
        data = self._graphql.post(
            """
            mutation ProvisionDataLabToMediaInsightsDcr($input: ProvisionDataLabInput!) {
                dataLab {
                    provisionDataLabToMediaInsightsDcr(input: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": {
                    "dataRoomId": data_room_id,
                    "dataLabId": data_lab_id,
                }
            },
        )
        return data["dataLab"]["provisionDataLabToMediaInsightsDcr"]["publishedDataLab"]

    def _deprovision_data_lab_from_ab_dcr(self, data_room_id: str) -> DataLabDefinition:
        """
        Deprovision a DataLab from an Audience Builder DCR.

        **Parameters**:
        - `data_room_id`: ID of the DCR to deprovision from.
        """
        data = self._graphql.post(
            """
            mutation DeprovisionDataLabFromMediaInsightsDcr($input: String!) {
                dataLab {
                    deprovisionDataLabFromMediaInsightsDcr(mediaInsightsDcrId: $input) {
                        publishedDataLab {
                            id
                            name
                            datasets {
                                name
                                dataset {
                                    id
                                    manifestHash
                                    name
                                }
                            }
                            usersDataset {
                                id
                                manifestHash
                                name
                            }
                            segmentsDataset {
                                id
                                manifestHash
                                name
                            }
                            demographicsDataset {
                                id
                                manifestHash
                                name
                            }
                            embeddingsDataset {
                                id
                                manifestHash
                                name
                            }
                            statistics
                            requireDemographicsDataset
                            requireEmbeddingsDataset
                            isValidated
                            numEmbeddings
                            matchingIdFormat
                            matchingIdHashingAlgorithm
                            validationComputeJobId
                            statisticsComputeJobId
                            jobsDriverAttestationHash
                            highLevelRepresentationAsString
                        }
                    }
                }
            }
            """,
            {
                "input": data_room_id,
            },
        )
        return data["dataLab"]["deprovisionDataLabFromMediaInsightsDcr"][
            "publishedDataLab"
        ]

    def _get_my_organization_users(self) -> List[OrganizationUser]:
        data = self._graphql.post(
            """
            query MyOrganizationUsers() {
                myself {
                    organization {
                        users {
                            nodes {
                                id
                                email
                            }
                        }
                    }
                }
            }
            """
        )
        return data["myself"]["organization"]["users"]["nodes"]



def create_client(
    user_email: str,
    api_token: str,
    *,
    client_id: str = DECENTRIQ_CLIENT_ID,
    api_host: str = DECENTRIQ_HOST,
    api_port: int = DECENTRIQ_PORT,
    api_use_tls: bool = DECENTRIQ_USE_TLS,
    request_timeout: Optional[int] = None,
    unsafe_disable_known_root_ca_check: bool = _DECENTRIQ_UNSAFE_DISABLE_KNOWN_ROOT_CA_CHECK,
) -> Client:
    """
    The primary way to create a `Client` object.

    **Parameters**:
    - `api_token`: An API token with which to authenticate oneself.
        The API token can be obtained in the user
        account settings in the Decentriq UI.
    - `user_email`: The email address of the user that generated the given API token.
    """

    enclave_api_token = api_token
    enclave_api_token_raw = urlsafe_b64decode(enclave_api_token)
    platform_api_token_raw = hashlib.sha256(enclave_api_token_raw).digest()
    platform_api_token = urlsafe_b64encode(platform_api_token_raw).decode("ascii")

    api = Api(
        platform_api_token,
        client_id,
        api_host,
        api_port,
        api_prefix="",
        use_tls=api_use_tls,
        timeout=request_timeout,
    )

    graphql = GqlClient(api, path=Endpoints.GRAPHQL)

    return Client(
        user_email,
        enclave_api_token,
        api,
        graphql,
        request_timeout=request_timeout,
        unsafe_disable_known_root_ca_check=unsafe_disable_known_root_ca_check,
    )


class BoundedExecutor:
    def __init__(self, bound, max_workers):
        self.executor = futures.ThreadPoolExecutor(max_workers=max_workers)
        self.semaphore = BoundedSemaphore(bound + max_workers)

    def submit(self, fn, *args, **kwargs):
        def done_callback(f):
            error = f.exception()
            if error:
                logger.error(f"Error in future: {error}")
            self.semaphore.release()

        self.semaphore.acquire()
        try:
            future = self.executor.submit(fn, *args, **kwargs)
        except:
            self.semaphore.release()
            raise
        else:
            future.add_done_callback(done_callback)
            return future

    def shutdown(self, wait=True):
        self.executor.shutdown(wait)

"""
Converts a generator of bytes into a read-only input stream. This takes ownership of the generator and will close it when the stream is closed.
"""
class GeneratorStream(io.RawIOBase):
    def __init__(self, generator, buffer_size=io.DEFAULT_BUFFER_SIZE):
        self.leftover = None
        self.has_next = True
        self.generator = generator
        self.buffer_size = buffer_size

    def readable(self):
        return True

    def readinto(self, output_buffer):
        if not self.has_next:
            return 0
        try:
            chunk = self.leftover or next(self.generator)
            read_size = min(len(output_buffer), len(chunk))
            output_buffer[:read_size] = chunk[:read_size]
            self.leftover = chunk[read_size:] if read_size < len(chunk) else None
            return read_size
        except StopIteration:
            self.has_next = False
            return 0

    def read(self, size=-1):
        if size == 0:
            return b""
        elif size < 0:
            buffers = []
            while self.has_next:
                buffer = bytearray(self.buffer_size)
                read_size = self.readinto(buffer)
                buffer = buffer[:read_size]
                buffers.append(buffer)
            return b"".join(buffers)
        else:
            buffer = bytearray(size)
            num_read = self.readinto(buffer)
            return bytes(buffer[:num_read])

    def close(self):
        self.generator.close()



from google.protobuf.json_format import MessageToDict

from ...proto import (
    ComputeNodeFormat,
    serialize_length_delimited,
)
from ..node import Node
from .proto import AzureBlobStorageWorkerConfiguration, ExportRole, SinkInput, ImportRole


class DataSourceAzureBlobStorage(Node):
    """
    Compute node that fetches a dataset from a Azure blob storage.
    """

    def __init__(
        self,
        name: str,
        credentials_dependency: str,
    ) -> None:
        config = AzureBlobStorageWorkerConfiguration(
            credentialsDependency=credentials_dependency, importRole=ImportRole()
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.azure-blob-storage-worker",
            dependencies=[credentials_dependency],
            output_format=ComputeNodeFormat.RAW,
        )


class DataSinkAzureBlobStorage(Node):
    """
    Compute node that exports data to Azure blob storage.
    """

    def __init__(
        self,
        name: str,
        credentials_dependency: str,
        input: SinkInput,
    ) -> None:
        config = AzureBlobStorageWorkerConfiguration(
            credentialsDependency=credentials_dependency,
            exportRole=ExportRole(input=input),
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.azure-blob-storage-worker",
            dependencies=[credentials_dependency, input.dependency],
            output_format=ComputeNodeFormat.RAW,
        )


from ...decoder import AzureBlobStorageWorkerDecoder

__all__ = [
    "DataSourceAzureBlobStorage",
    "DataSinkAzureBlobStorage",
    "AzureBlobStorageWorkerDecoder",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: amazon_amc.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'amazon_amc.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61mazon_amc.proto\x12\namazon_amc\"\x1a\n\nSingleFile\x12\x0c\n\x04name\x18\x01 \x01(\t\"\t\n\x07RawFile\"D\n\x07ZipFile\x12,\n\nsingleFile\x18\x01 \x01(\x0b\x32\x16.amazon_amc.SingleFileH\x00\x42\x0b\n\tselection\"}\n\tDataInput\x12\x12\n\ndependency\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\"\n\x03raw\x18\x03 \x01(\x0b\x32\x13.amazon_amc.RawFileH\x00\x12\"\n\x03zip\x18\x04 \x01(\x0b\x32\x13.amazon_amc.ZipFileH\x00\x42\x06\n\x04\x66ile\"\x0f\n\rUnHashedEmail\"\x13\n\x11HashedEmailSha256\"J\n\x0eTargetResource\x12\x16\n\x0c\x63onnectionId\x18\x01 \x01(\tH\x00\x12\x16\n\x0c\x61\x64vertiserId\x18\x02 \x01(\tH\x00\x42\x08\n\x06target\"\xc9\x02\n\x1c\x41mazonAmcWorkerConfiguration\x12$\n\x05input\x18\x01 \x01(\x0b\x32\x15.amazon_amc.DataInput\x12\x1d\n\x15\x63redentialsDependency\x18\x02 \x01(\t\x12\x32\n\x0etargetResource\x18\x03 \x01(\x0b\x32\x1a.amazon_amc.TargetResource\x12\x12\n\naudienceId\x18\x04 \x01(\t\x12\x14\n\x0c\x61udienceName\x18\x05 \x01(\t\x12 \n\x13\x61udienceDescription\x18\x06 \x01(\tH\x00\x88\x01\x01\x12$\n\x07\x63ountry\x18\x07 \x01(\x0e\x32\x13.amazon_amc.Country\x12&\n\x08idFormat\x18\x08 \x01(\x0e\x32\x14.amazon_amc.IdFormatB\x16\n\x14_audienceDescription*a\n\x08IdFormat\x12\x12\n\x0eUNHASHED_EMAIL\x10\x00\x12\x17\n\x13HASHED_EMAIL_SHA256\x10\x01\x12\x0c\n\x08PHONE_NR\x10\x02\x12\x1a\n\x16HASHED_PHONE_NR_SHA256\x10\x03*\xc8\x02\n\x07\x43ountry\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x18\n\x14UNITED_ARAB_EMIRATES\x10\x01\x12\r\n\tAUSTRALIA\x10\x02\x12\x0b\n\x07\x42\x45LGIUM\x10\x03\x12\n\n\x06\x42RAZIL\x10\x04\x12\n\n\x06\x43\x41NADA\x10\x05\x12\x0b\n\x07GERMANY\x10\x06\x12\t\n\x05\x45GYPT\x10\x07\x12\t\n\x05SPAIN\x10\x08\x12\n\n\x06\x46RANCE\x10\t\x12\x12\n\x0eUNITED_KINGDOM\x10\n\x12\t\n\x05INDIA\x10\x0b\x12\t\n\x05ITALY\x10\x0c\x12\t\n\x05JAPAN\x10\r\x12\n\n\x06MEXICO\x10\x0e\x12\x0f\n\x0bNETHERLANDS\x10\x0f\x12\n\n\x06POLAND\x10\x10\x12\x10\n\x0cSAUDI_ARABIA\x10\x11\x12\n\n\x06SWEDEN\x10\x12\x12\r\n\tSINGAPORE\x10\x13\x12\n\n\x06TURKEY\x10\x14\x12\x1c\n\x18UNITED_STATES_OF_AMERICA\x10\x15\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'amazon_amc_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_IDFORMAT']._serialized_start=714
  _globals['_IDFORMAT']._serialized_end=811
  _globals['_COUNTRY']._serialized_start=814
  _globals['_COUNTRY']._serialized_end=1142
  _globals['_SINGLEFILE']._serialized_start=32
  _globals['_SINGLEFILE']._serialized_end=58
  _globals['_RAWFILE']._serialized_start=60
  _globals['_RAWFILE']._serialized_end=69
  _globals['_ZIPFILE']._serialized_start=71
  _globals['_ZIPFILE']._serialized_end=139
  _globals['_DATAINPUT']._serialized_start=141
  _globals['_DATAINPUT']._serialized_end=266
  _globals['_UNHASHEDEMAIL']._serialized_start=268
  _globals['_UNHASHEDEMAIL']._serialized_end=283
  _globals['_HASHEDEMAILSHA256']._serialized_start=285
  _globals['_HASHEDEMAILSHA256']._serialized_end=304
  _globals['_TARGETRESOURCE']._serialized_start=306
  _globals['_TARGETRESOURCE']._serialized_end=380
  _globals['_AMAZONAMCWORKERCONFIGURATION']._serialized_start=383
  _globals['_AMAZONAMCWORKERCONFIGURATION']._serialized_end=712
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import json
from typing import Dict, Any, List, Optional, TYPE_CHECKING

from .computations import Computation
from ..session import Session
from ..storage import Key
from decentriq_dcr_compiler.schemas import RequestedAudiencePayload
from decentriq_dcr_compiler.schemas import (
    AbMediaRequest,
    RequestedAudiencePayload,
)

if TYPE_CHECKING:
    from ..client import Client


class PublishAudiencesDataset(Computation):
    def __init__(
        self, dcr_id: str, dataset_id: str, key: Key, client: Client, session: Session
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.dataset_id = dataset_id
        self.key = key

    def node_id(self) -> str:
        pass

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="publishAudiencesDataset")


class GetAudiencesForAdvertiserComputation(Computation):
    def __init__(self, dcr_id: str, client: Client, session: Session) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "get_audiences_for_advertiser"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="getAudiencesForAdvertiser")

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "audiences.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class RunCreateAudienceUserList(Computation):
    def __init__(self, dcr_id: str, client: Client, session: Session) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "run_create_audience_user_list"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="runCreateAudienceUserList")


class EstimateAudienceSizeForAdvertiserComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience

    def node_id(self) -> str:
        return "estimate_audience_size_for_advertiser"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="estimateAudienceSizeForAdvertiser",
            generate_audience=self.generate_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "audience_size.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class EstimateAudienceSizeForAdvertiserLalComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
        lal_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience
        self.lal_audience = lal_audience

    def node_id(self) -> str:
        return "estimate_audience_size_for_advertiser_lal"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="estimateAudienceSizeForAdvertiserLal",
            generate_audience=self.generate_audience,
            lal_audience=self.lal_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "audience_size.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class GetAudienceUserListForAdvertiserComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience

    def node_id(self) -> str:
        return "get_audience_user_list_for_advertiser"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="getAudienceUserListForAdvertiser",
            generate_audience=self.generate_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> List[str]:
        result = super().get_results_str_from_zip(
            "audience_users.csv", interval=interval, timeout=timeout
        )
        return [line for line in result.split("\n") if line]


class GetAudienceUserListForAdvertiserLalComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
        lal_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience
        self.lal_audience = lal_audience

    def node_id(self) -> str:
        return "get_audience_user_list_for_advertiser_lal"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="getAudienceUserListForAdvertiserLal",
            generate_audience=self.generate_audience,
            lal_audience=self.lal_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> List[str]:
        result = super().get_results_str_from_zip(
            "audience_users.csv", interval=interval, timeout=timeout
        )
        return [line for line in result.split("\n") if line]


class PublishAudiencesJson(Computation):
    def __init__(
        self,
        dcr_id: str,
        manifest_hash: str,
        encryption_key_hex: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.manifest_hash = manifest_hash
        self.encryption_key_hex = encryption_key_hex

    def node_id(self) -> str:
        return "audiences.json"

    def run(self) -> None:
        """
        Publish the `audiences.json` dataset.
        """
        request_type = "publishAudiencesJson"
        request = AbMediaRequest.model_validate(
            {
                request_type: {
                    "dataRoomIdHex": self.dcr_id,
                    "datasetHashHex": self.manifest_hash,
                    "encryptionKeyHex": self.encryption_key_hex,
                    "scopeIdHex": self.client._ensure_dcr_data_scope(self.dcr_id),
                },
            }
        )
        response = self.send_request(request).model_dump()


class RetrievePublishedDatasetsComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        pass

    def run_and_get_results(self) -> Dict[str, Any]:
        """
        Retrieve all datasets published to the DCR.
        """
        request_type = "retrievePublishedDatasets"
        request = AbMediaRequest.model_validate(
            {
                request_type: {
                    "dataRoomIdHex": self.dcr_id,
                },
            }
        )
        response = self.send_request(request).model_dump()
        return response["retrievePublishedDatasets"]


class GetLookalikeAudienceStatisticsComputation(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
        generate_audience: RequestedAudiencePayload,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)
        self.generate_audience = generate_audience

    def node_id(self) -> str:
        return "get_lookalike_audience_statistics_id"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(
            request_type="getLookalikeAudienceStatistics",
            generate_audience=self.generate_audience,
        )

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "lookalike_audience.json", interval=interval, timeout=timeout
        )
        results_json = json.loads(results)

        # Only include the necessary information to the user.
        quality_metrics = {"quality": results_json["quality"]}
        all_roc_curve_keys = quality_metrics["quality"]["roc_curve"].keys()
        keep_roc_curve_keys = ["auc", "fpr", "tpr"]
        drop_keys = list(set(all_roc_curve_keys) - set(keep_roc_curve_keys))
        [quality_metrics["quality"]["roc_curve"].pop(m, None) for m in drop_keys]
        return quality_metrics


class GetDataAttributesComputation(Computation):
    def __init__(self, dcr_id: str, client: Client, session: Session) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "get_data_attributes"

    def run(self) -> None:
        """
        Run the computation.
        """
        super().run(request_type="getDataAttributes")

    def get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> Dict[str, Any]:
        results = super().get_results_str_from_zip(
            "attributes.json", interval=interval, timeout=timeout
        )
        return json.loads(results)


class GetAudiencesValidationReport(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "audiences_validation_report"

    def run(self) -> None:
        super().run(
            request_type="getAudiencesValidationReport",
        )

    def run_and_get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> dict[str, Any]:
        self.run()
        report_str = super().get_results_str_from_zip(
            "validation-report.json", interval=interval, timeout=timeout
        )
        report = json.loads(report_str)
        return {
            "audiences": report
        }


class ComputeOverlapStatistics(Computation):
    def __init__(
        self,
        dcr_id: str,
        client: Client,
        session: Session,
    ) -> None:
        super().__init__(dcr_id=dcr_id, client=client, session=session)

    def node_id(self) -> str:
        return "compute_overlap_statistics"

    def run(self) -> None:
        super().run(
            request_type="computeOverlapStatistics",
        )

    def run_and_get_results(
        self, interval: int = 5, timeout: Optional[int] = None
    ) -> dict[str, Any]:
        self.run()
        statistics_str = super().get_results_str_from_zip(
            "overlap.json", interval=interval, timeout=timeout
        )
        statistics = json.loads(statistics_str)
        return statistics
